/*************************************************************************************
file name: config.h
author: qingzhe huang
date: Jul. 21, 2007

purpose: This is used to define all those configuration other than system API.
1. Geometry data structure type definition.
2. MPI running configure:
 TesterCount+RenderCount+1 = "size" of total nodes running for MPI
[TesterHostStartIndex, TesterHostStartIndex+TesterCount]: Tester Index Range
[RenderHostStartIndex, RenderHosterStartIndex+RenderCount]: Render Index Range

a)
MasterIndex can be any index between 0 to size-1 as long as it doesn't fall within either 
Tester index or Render index.
b) 
Tester Index Range and Render Index Range cannot overlap.

3. octreeLoader configuration:
a) Different system may have different physical memory and when physical memory is 
not enough "file mapping" will fail. So, constant "ThresholdRate" setup a threshold 
limit for "unmap" starts to free unused mapping by checking if it appears in past
"MaxCoherentFrameNumber" frames. If you have large physical memory and you want better
performance, you can adjust both these number to bigger.

4. Linux user should best check "PageSize" by "sysconf" because I just hardcoded 4k. 
5. Originally I thought it might help to improve "depth-buffer-precision" by using
double-precision floating number instead of default single-precision. Later I think
OpenGL might internally use single-precision. This is not a sure thing, but for performance
purpose I stick to single. This is why I have macro of "USE_DOUBLE_PRECISION".
6. OpenGL maximumly support 8 lights, and the "MaxLightNumber" should not exceed 8. Of
course to speed up, you can set it smaller. However, if you want to change this number, you
may need to modify ininitialized light color, position manually in "renderEngine.cpp".
*************************************************************************************/

#ifndef CONFIG_H
#define CONFIG_H
/////////////////////////////////////
//define various constants, datatypes, macros etc
#include "syscall.h"
#include "coreConfig.h"

#ifndef LINUX_VERSION

#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "OpenGL32.lib")
//#pragma comment(lib, "wmpi2.lib")
//#pragma comment(lib, "wmpi2cppbindings.lib")
const unsigned int PageSize=64*1024;

#else
//this must be checked by "sysconf", but I just hardcoded for the time being.
const unsigned int PageSize=4*1024;
#endif


/////////////////////////////////////////
//octree data types
#define CHECK_NEW(x) if (x==NULL){printf("memory alloc failed\n"); exit(1);}
#define MIN(x, y) ((x<y)?(x):(y))
#define MAX(x, y) ((x>y)?(x):(y))


typedef unsigned char ubyte;
typedef unsigned short ushort;

//#define USE_DOUBLE_PRECISION 1

typedef float FVector[3];
typedef double DVector[3];
typedef ubyte CVector[4];

//always use VerVector
#ifdef USE_DOUBLE_PRECISION
	#define VVector DVector
	#define VType double
#else
	#define VVector FVector
	#define VType float
#endif

/////////////////////////////////////////////
//octree constants

const VType Epsilon=0.001;
const VType MinimumEdgeLength=0.000001;

const unsigned int DefaultMaxLeafNodeVertexNumber=1024;
const unsigned int MaxPossibleDepth=30;

const unsigned int MaxNeighbourCount=26;



/////////////////////////////////////////////////
//constants for octreeloader

const unsigned int MaxVertexCapacity=64*1024;
const unsigned int MaxFaceCapacity=64*1024;
const unsigned int MaxCapacity=2*1024*1024*1024/MaxFaceCapacity;
const VType ThresholdRate=200;
const unsigned int MaxCoherentFrameNumber=2;
const unsigned int MaxPossibleNodeCollectable=64*1024;
const unsigned int DisplayOctreeLeafNodeBatchNumber = 100;
//////////////////////////////////////////////////////
//master's data structure
/*
const unsigned int MasterIndex=1131;
const unsigned int RenderHostStartIndex=1116;
const unsigned int TesterHostStartIndex=0;
const unsigned int TesterCount=128;//default number
const unsigned int TesterHostCount=TesterCount;
const unsigned int DisplayerIndex=1131;
*/
//const unsigned int RenderCount=RenderHostCount;

////////////////////////////////////////
const unsigned int ScreenSize =400;
const unsigned int ColorBufferSize=ScreenSize*ScreenSize;
const unsigned int ColorBufferByteSize=ColorBufferSize*4;
const unsigned int DepthBufferSize=ColorBufferSize;
const unsigned int DepthBufferByteSize=sizeof(float)*DepthBufferSize;

//this is used to make sure compression will not overbust buffer
const unsigned int VisibleTableBufferRate=2;
const unsigned int VisibleTableSize=ColorBufferSize*VisibleTableBufferRate;

///////////////////////////////////////////////////////
//render engine
const unsigned int scrW=ScreenSize;
const unsigned int scrH=ScreenSize;
const unsigned int MaxLightNumber=8;

/////////////////////////////////////////////////
//constants for displayer
const unsigned int DepthBufferAtomWidth=200;
const unsigned int DepthBufferAtomSize=DepthBufferAtomWidth*DepthBufferAtomWidth;
const unsigned int DepthBufferSnapShotWidth=ScreenSize/DepthBufferAtomWidth;
const unsigned int DepthBufferSnapShotNumber=DepthBufferSnapShotWidth*DepthBufferSnapShotWidth;
const unsigned int ColorBufferAtomSize=DepthBufferAtomSize*4;
/////////////////////////////////////////
//profiling
const int START_FRAME_NUMBER=25;
const int QUIT_FRAME_NUMBER=150;
const int FRAME_NUMBER_OFFSET=25;

//static const VVector StartingOffset={-0.20, 0, -0.20};
static const VVector StartingOffset={-0.10, 0, -0.10};
//static const VVector StartingOffset={-0.15, 0, -0.15};//expand5
//static const VVector StartingOffset={-0.05, -0.05, -0.05};
static const float LightPositionOffset=15;//powerplant
//static const float LightPositionOffset=105;//statue
static const VVector LucyStartingOffset={-878,-188,621};//lucy
#endif
