
#include <mpi.h>
#include "config.h"
#include "dataStructure.h"
#include <cstdio>
#include <cstdlib>
#include <GL/glut.h>
#include "except.h"
#include <assert.h>

extern FILE* dumpPtr;



char* MPIContext::renderContextBuffer=NULL;

MPIContext mpiContext;

#ifdef PROFILE_ALL
bool startTracing=false;
#endif


void PROFILE_BEGIN(ProfileEnum item)
{
#ifdef  PROFILING
        mpiContext.profiler.begin(item);
#endif
}

void PROFILE_END(ProfileEnum item)
{
#ifdef  PROFILING
        mpiContext.profiler.end(item);
#endif
}

void PROFILE_DISPLAY()
{
#ifdef  PROFILING
        mpiContext.profiler.display();
#endif
}

void PROFILE_PRINT()
{
#ifdef  PROFILING
        mpiContext.profiler.print();
	mpiContext.profiler.printfps();
#endif
}

void Profiler::init(int myStartFrame, int myEndFrame)
{
	memset(last, 0, sizeof(long)*ProfileEnumCount);
	memset(total, 0, sizeof(long)*ProfileEnumCount);
	startFrameNumber=myStartFrame;
	endFrameNumber=myEndFrame;
	frameNumber=0;
}

inline void Profiler::begin(ProfileEnum item)
{
	last[item]=GenericGetProfileTime();
	if (item==ALL)
	{
		frameNumber++;
	}
}

inline void Profiler::end(ProfileEnum item)
{
	current[item]=GenericGetProfileTime();
	if (frameNumber>=startFrameNumber&&frameNumber<endFrameNumber)
	{
		total[item]+=current[item]-last[item];
	}
}

void Profiler::display()
{
	printf("Node[%s]name[%s]frameNumber[%d]:\n", NodeName[mpiContext.myNodeStatus], mpiContext.nodeName, endFrameNumber-startFrameNumber);
	for (int i=0; i<ProfileEnumCount; i++)
	{
		if (total[i]==0)
		{
			continue;
		}
		printf("%s[%d]\n", ProfileItemName[i], total[i]);
		if (i<ProfileEnumCount-1&&total[ProfileEnumCount-1]!=0)
		{
			printf("%s/total=%f\%\n", ProfileItemName[i], 
				(float)(total[i])/(float)(total[ProfileEnumCount-1])*100.0);
		}
	}
}

void Profiler::printfps()
{
	printf("average fps [%f]\n", (QUIT_FRAME_NUMBER-START_FRAME_NUMBER-FRAME_NUMBER_OFFSET)*1000000.0/total[ALL]);
}

void Profiler::print()
{
	fprintf(filePtr, "Node[%s]name[%s]frameNumber[%d]:\n", NodeName[mpiContext.myNodeStatus], 
		mpiContext.nodeName, endFrameNumber-startFrameNumber);
	for (int i=0; i<ProfileEnumCount; i++)
	{
		if (total[i]==0)
		{
			continue;
		}
		fprintf(filePtr, "[%s][%d]\n", ProfileItemName[i], total[i]);
/*
		if (i<ProfileEnumCount-1&&total[ProfileEnumCount-1]!=0)
		{
			fprintf(filePtr, "[%s/total]=[%f\%]\n", ProfileItemName[i], 
				(float)(total[i])/(float)(total[ProfileEnumCount-1])*100.0);
		}
*/
	}
	//fclose(filePtr);
	//filePtr=NULL;
}
	

void MPIContext::init()
{
	int i, rankArraySize;
	MPI_Group commGroup;
	DEBUG_INFO("mpiContext init...\n");
	MPI_Comm_size(MPI_COMM_WORLD, &mySize);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	renderContextBuffer=new char[MaxRenderContextBufferSize];
	CHECK_NEW(renderContextBuffer);
	renderContext=(RenderContext*)renderContextBuffer;

	pendingKey=0;
	startMPI=false;

	renderContext->indexCount=0;
	renderContext->frameNumber=0;
	renderContext->key=0;//default key

	renderShareArray=testerShareArray=NULL;
	DEBUG_INFO("before node status...\n");
	if (myRank==MasterIndex)
	{
		myNodeStatus=MASTER;
		contextRequests=new MPI_Request[mySize];
		CHECK_NEW(contextRequests);	
		for (i=0; i<mySize; i++)
		{
			contextRequests[i]=MPI_REQUEST_NULL;		
		}


	}
#ifdef	USING_SEPARATE_DISPLAYER
	if (myRank==DisplayerIndex)
	{
		myNodeStatus=DISPLAYER;
		contextRequests=new MPI_Request;
		CHECK_NEW(contextRequests);
		contextRequests[0]=MPI_REQUEST_NULL;
		//mpiContext.key='b';
	}
#endif
	if (myRank>=TesterHostStartIndex&&myRank<TesterHostStartIndex+TesterCount)
	{
		myNodeStatus=TESTER;
		contextRequests=new MPI_Request;
		CHECK_NEW(contextRequests);
		contextRequests[0]=MPI_REQUEST_NULL;
	}
	if (myRank>=RenderHostStartIndex&&myRank<RenderHostStartIndex+RenderCount)
	{
		myNodeStatus=RENDER;
		contextRequests=new MPI_Request;
		CHECK_NEW(contextRequests);
		contextRequests[0]=MPI_REQUEST_NULL;
	}
	profiler.init(START_FRAME_NUMBER+FRAME_NUMBER_OFFSET, QUIT_FRAME_NUMBER);
	DEBUG_INFO("after node status...\n");
}



//masterContext will guarantee renderContext->startMpi be modified to be true BEFORE "sendRenderContext"
//is called.
#ifdef 	USING_BROADCAST
void MPIContext::sendRenderContext()
{
	int i;
	int broadcastSize;
	//to make sure the "very" first render context is still sending by "synchronized"
	//so that the receiving party can receive it immediately
#ifdef USING_SEPARATE_DISPLAYER
	if (renderContext->frameNumber==QUIT_FRAME_NUMBER)
	{
		pendingKey='q';
	}
	else
	{
		if (renderContext->frameNumber==START_FRAME_NUMBER)
		{
			pendingKey='b';
		}
	}
#endif
#ifdef	PROFILING
	if (renderContext->frameNumber==START_FRAME_NUMBER)
	{
		printf("**************start profiling...........*************\n");
		pendingKey='b';
	}
#endif


	if (pendingKey!=0)
	{
		renderContext->key=pendingKey;
	}
//////////////////////////////////////////////
#ifdef PROFILE_ALL
	if (!startTracing)
	{
		startTracing=true;
		MPIHP_Trace_on();
	}
#endif
/////////////////////////////////////////////
	//printf("master is about to send render context no. %d to host %d\n", renderContext->frameNumber, i);
	//i is the rank
	//SAFE_CALL(MPI_Bcast(renderContext, MaxRenderContextBufferSize, MPI_UNSIGNED_CHAR, MasterIndex, MPI_COMM_WORLD));				
	SAFE_CALL(MPI_Bcast(renderContext, MaxRenderContextBufferSize, MPI_UNSIGNED_CHAR, 0, broadcastComm));				
			//else index remain the same

	//this will guarantee "master" send first notifying message with "synchoronized method"
	if (renderContext->key=='B'||renderContext->key=='b')
	{
		startMPI=true;
	}

	if (renderContext->key=='q'||renderContext->key=='Q')
	{
/////////////////////////////////////////////
#ifdef PROFILE_ALL
		if (startTracing)
		{
			MPIHP_Trace_off();
		}
#endif
/////////////////////////////////////////////
		quit();
	}
}
#else//non_broadcast
void MPIContext::sendRenderContext()
{
	int i;
	//to make sure the "very" first render context is still sending by "synchronized"
	//so that the receiving party can receive it immediately
#ifdef USING_SEPARATE_DISPLAYER
	if (renderContext->frameNumber==QUIT_FRAME_NUMBER)
	{
		pendingKey='q';
	}
	else
	{
		if (renderContext->frameNumber==START_FRAME_NUMBER)
		{
			pendingKey='b';
		}
	}
#endif
#ifdef	PROFILING
	if (renderContext->frameNumber==START_FRAME_NUMBER)
	{
		printf("**************start profiling...........*************\n");
		pendingKey='b';
	}
#endif

//in HP, our displayer is visible, but master or control is invisible
	if (pendingKey!=0)
	{
		renderContext->key=pendingKey;
	}

	if (startMPI)
	{
//////////////////////////////////////////////
#ifdef PROFILE_ALL
		if (!startTracing)
		{
			startTracing=true;
			MPIHP_Trace_on();
		}
#endif
/////////////////////////////////////////////
		for (i=0; i<mySize; i++)
		{	
#ifndef	USING_SYNCHRONIZATION
			if (i==DisplayerIndex)
			{
				continue;
			}
#endif
			if (i!=MasterIndex)
			{
				//printf("master is about to send render context no. %d to host %d\n", renderContext->frameNumber, i);
				//i is the rank
				SAFE_CALL(MPI_Issend(renderContext, renderContextSize, MPI_UNSIGNED_CHAR, i, RENDER_CONTEXT, 
					MPI_COMM_WORLD, contextRequests+i));				
			}
			//else index remain the same
		}
	}
	else
	{
		for (i=0; i<mySize; i++)
		{	
#ifndef	USING_SYNCHRONIZATION
			if (i==DisplayerIndex)
			{
				continue;
			}
#endif
			if (i!=MasterIndex)
			{
				//printf("master is about to send render context no. %d to host %d\n", renderContext->frameNumber, i);
				//i is the rank
				SAFE_CALL(MPI_Ssend(renderContext, renderContextSize, MPI_UNSIGNED_CHAR, i, RENDER_CONTEXT, 
					MPI_COMM_WORLD));					
			}			
			//else index remain the same
		}
		//printf("master finished sending render context of frame number %d\n", renderContext->frameNumber);		
	}

	//this will guarantee "master" send first notifying message with "synchoronized method"
	if (renderContext->key=='B'||renderContext->key=='b')
	{
		startMPI=true;
	}
////////////////////


}
#endif

void MPIContext::quit()
{
	printf("rank[%d]name[%s]node[%s] abort on frame [%d]\n", myRank, NodeName[myNodeStatus], nodeName, renderContext->frameNumber);
/*
	if (dumpPtr!=NULL)
	{
		fclose(dumpPtr);
		dumpPtr=NULL;
	}
*/

	MPI_Finalize();
#ifdef	 PROFILING
	//printf("[%s] print profile\n", nodeName);
	profiler.print();
	fclose(profiler.filePtr);
#endif
	exit(0);
}

//we have to reset the key value, otherwise it is not an event, but a status!!!!
void MPIContext::testRenderContextSending()
{
#ifndef	USING_BROADCAST
	if (startMPI)
	{
		SAFE_CALL(MPI_Waitall(mySize, contextRequests, MPI_STATUSES_IGNORE));
	}
	//printf("master finished sending render context of frame number %d\n", renderContext->frameNumber);
#endif
	//quit condition
	if (renderContext->key=='q'||renderContext->key=='Q')
	{		
		//SAFE_CALL(MPI_Waitall(mySize, contextRequests, MPI_STATUSES_IGNORE));
/////////////////////////////////////////////
#ifdef PROFILE_ALL
		if (startTracing)
		{
			MPIHP_Trace_off();
		}
#endif
/////////////////////////////////////////////
		quit();
	}


	renderContext->frameNumber++;

	if (pendingKey==renderContext->key)
	{
		//this measn the data is already sent
		pendingKey=renderContext->key=0;
	}
}

void MPIContext::testRenderContextRecving()
{
#ifndef	 USING_BROADCAST
	if (startMPI)
	{
		SAFE_CALL(MPI_Wait(contextRequests, MPI_STATUS_IGNORE));
	}
	//printf("host receives rendercontext of frame no. %d\n", renderContext->frameNumber);
#endif


	if (renderContext->key=='q'||renderContext->key=='Q')
	{	
/////////////////////////////////////////////
#ifdef PROFILE_ALL
		if (startTracing)
		{
			MPIHP_Trace_off();
		}
#endif
/////////////////////////////////////////////
		//renderTestVisiblePixelInfoRecving();
		//renderTestPixelDataSending();
		quit();
	}
///////////////////////////////////////////////////////////////////////

	renderShareArray=(int*)(renderContextBuffer+sizeof(RenderContext));
	testerShareArray=(int*)(renderContextBuffer+sizeof(RenderContext)+sizeof(int)*renderContext->indexCount);
	if (renderContext->key=='b'||renderContext->key=='B')
	{
//////////////////////////////////////////////
#ifdef PROFILE_ALL
		if (!startTracing)
		{
			startTracing=true;
			MPIHP_Trace_on();
		}
#endif
/////////////////////////////////////////////
		startMPI=true;
	}
}


void MPIContext::recvRenderContext()
{

#ifdef	USING_BROADCAST
	SAFE_CALL(MPI_Bcast(renderContext, MaxRenderContextBufferSize, MPI_UNSIGNED_CHAR, 0, broadcastComm));
#else
	if (startMPI)
	{
		SAFE_CALL(MPI_Irecv(renderContext, MaxRenderContextBufferSize, MPI_UNSIGNED_CHAR, MasterIndex, RENDER_CONTEXT, 
			MPI_COMM_WORLD, contextRequests))
	}
	else
	{
		SAFE_CALL(MPI_Recv(renderContext, MaxRenderContextBufferSize, MPI_UNSIGNED_CHAR, MasterIndex, RENDER_CONTEXT, 
			MPI_COMM_WORLD, MPI_STATUS_IGNORE))
	}
#endif
}

void getLastErrorMsg()
{
	char msgBuf[256];
	int length;
	if (MPI_error_code!=MPI_SUCCESS)
	{
		MPI_Error_string(MPI_error_code, msgBuf, &length);
		fprintf(dumpPtr, "the mpi error is:%s\n", msgBuf);
	
	}
}

//a simple run-length compression, return the byte size of compressed buffer, -1 indicating buffer overflow
int compressBuffer(ubyte* source, int length, ubyte* target, int maxBufferSize)
{
        int i;
	int count=0;
	assert(length>0);
        ubyte current[2];//treat ushort as if ubyte[2];
        current[0]=source[0];
        current[1]=1;
        for (i=1; i<length; i++)
        {
                if (current[0]==source[i])
                {
                        if (current[1]<255)
                        {
                                current[1]++;
                        }
                        else
                        {
                                //the counter is full, we need to write
                               	memcpy(target+2*count, current, 2);
                       		count++;
                          	
				current[1]=1;//restart
                                //we don't change current[0]
                        }
                }
                else
                {
                        //start new one after we write
                        memcpy(target+2*count, current, 2);
                       	count++;
                        current[0]=source[i];
                        current[1]=1;
                }
        }
        //finally we still have to write the last one
	//lazy checking, only do it once at last, so I don't guarantee buffer is NOT overwritten.
	assert(count*2<maxBufferSize);
       	memcpy(target+2*count, current, 2);
       	count++;
	//printf("compressed size %u, rate %f\%\n", count*2, (float)count*2.0*100.0/(float)maxBufferSize);
        return count*2;
}




int decompressBuffer(ubyte* source, int length, ubyte* target, int maxBufferSize)
{
        int i, offset=0;
        for (i=0; i<length; i+=2)
        {
                memset(target+offset, source[i], source[i+1]);
                offset+=source[i+1];
        }
	assert(offset<=maxBufferSize);

	//lazy checking, so, maybe the buffer is already overwritten. no duty for me.
	return offset;
}


