/*****************************************************************************************
author: qingzhe huang
date: Jul.31, 2007
1. The trick of "MPIContext" structure is that it is a structure with variable length. i.e.
For each frame, you don't know how many "octree node index" are used to represent total model.
This is a bit inconvenient in networking cause you always expect to send something fixed size.
(Maybe this is not the appropriate term. Anyway it is inconvenient for other modules.)
So, within MPIContext structure I set another structure to give information of variable length
index set array. It is "RenderContext" structure.
2.DepthBufferAtomSize should be checked such that ScreenSize is multiple times of its value.
Otherwise screen is not properly divided into "tiles".
  



****************************************************************************************/

#ifndef DATASTRUCTURE_H
#define DATASTRUCTURE_H

#include "config.h"

#include <mpi.h>

void getLastErrorMsg();

#ifdef DEBUGGING
#define DEBUG_INFO(x) printf(x);
#else
//this is a tricky question, if people use lazy "if (x) DEBUG(y) z;"
//instead of safer "if (x) { DEBUG(y) } z;"
#define DEBUG_INFO(x)   ;
#endif


static int MPI_error_code=MPI_SUCCESS;
#ifdef DEBUGGING
#define SAFE_CALL(x) if ((MPI_error_code=x)!=MPI_SUCCESS) getLastErrorMsg();
#else
#define SAFE_CALL(x) x;
#endif


//const int START_FRAME_NUMBER=30;
//const int QUIT_FRAME_NUMBER=1050;
const int RENDER_CONTEXT=13000;
const int VISIBLE_PIXEL_INFO=14000;
const int PIXEL_DATA=15000;
const int DEPTH_BUFFER_SNAP_SHOT=16000;
const int BINARY_SWAP_DEPTH_BUFFER=17000;
const int BINARY_SWAP_COLOR_BUFFER=19000;
const int BINARY_SWAP_VISIBLE_TABLE=18000;

struct RenderContext
{
	unsigned char key;
	unsigned char reserved0;
	unsigned char reserved1;
	unsigned char reserved2;
	int frameNumber;
	//int menuChoice;

	//VType fMat[16];	
	int indexCount;
	int renderShare[RenderCount];
	int testerShare[TesterCount];
};


const int MaxRenderContextBufferSize=sizeof(RenderContext)+MaxPossibleNodeCollectable*2*sizeof(int);

enum NodeStatus
{
	MASTER, RENDER, TESTER, DISPLAYER
};

const static char* NodeName[4]=
{
	"Master Node",
	"Render Node",
	"Tester Node",
	"Displayer Node"
};
enum ProfileEnum
{
	CM=0, CT=1, CR=2, CD=3, BMTRD=4, BTRD=5, BRD=6, BT=7,  ALL=8
};
const int ProfileEnumCount=9;

static char* ProfileItemName[ProfileEnumCount]=
{
	"computation of Master",
	"computation of Tester",
	"computation of Render",
	"computation of Displayer",
	"communication from Master to Tester, Render and Displayer",
 	"communication from Tester to Render, Displayer",
	"communication from Render to Displayer",
	"communication within Tester",
	"whole system"
};

struct  Profiler
{
	FILE* filePtr;
	int frameNumber;
	long int last[ProfileEnumCount];
	long int current[ProfileEnumCount];
	long int total[ProfileEnumCount];
	int startFrameNumber;
	int endFrameNumber;
	void init(int myStartFrame, int myEndFrame);
	void begin(ProfileEnum item);
	void end(ProfileEnum item);
	void print();
	void display();
	void printfps();
};	



//common for every node
//keyboard should not modify directly the "renderContext->startMPI" flag. 
//Instead modify "keyboardStartMPIFlag".
//similarly keyboard should not modify "
struct MPIContext
{
	int myRank;
	int mySize;
	NodeStatus myNodeStatus;
	char nodeName[32];
	bool startMPI;
	unsigned char pendingKey; //

	static char* renderContextBuffer;

	MPI_Request* contextRequests;
	RenderContext* renderContext;
	int renderContextSize;
	int* renderShareArray;
	int* testerShareArray;

	MPI_Comm broadcastComm;
	MPI_Group broadcastGroup;
	int* rankArray;

	Profiler profiler;
	void recvRenderContext();
	void testRenderContextSending();
	void sendRenderContext();
	void testRenderContextRecving();

	void quit();
	void init();
};

//static MPIContext mpiContext;
int decompressBuffer(ubyte* source, int length, ubyte* target, int maxBufferSize);
int compressBuffer(ubyte* source, int length, ubyte* target, int maxBufferSize);


void PROFILE_BEGIN(ProfileEnum item);

void PROFILE_END(ProfileEnum item);

void PROFILE_DISPLAY();
void PROFILE_PRINT();

#endif
