#include "config.h"
#include "dataStructure.h"
#include "binarySwap.h"
#include "master.h"
#include <cstdlib>
#include <cstdio>
#include <mpi.h>
#include <GL/glut.h>
#include <assert.h>
#include <GL/glut.h>
#include <unistd.h>


#ifdef DEBUGGING
#ifdef LINUX_VERSION
#include <syslog.h>

#endif
#endif

char* TesterContext::windowTitle="Tester Node";
extern MPIContext mpiContext;
extern FILE* dumpPtr;
////////////////////////////
//convenient alias
//float* myRecvDepthBuffer;
//float* myDepthBuffer;
//float* myExchangeDepthBuffer;
//float* myDepthBufferSnapShots;


void DepthBufferExchangeDescriptor::display()
{
	int i;
	printf("exchangde descriptor: myIndex:%d, testerCount:%d:", myIndex, testerCount);
	for (i=0; i<testerCount; i++)
	{
		printf("[%d]", testers[i]);
	}
	printf("\n");
}


//assume TileCount and TesterCount both being power of 2!!!!!!!!!
void BinarySwapDescriptor::init(int testerIndex)
{
	myTesterIndex=testerIndex;
	mask=0xFFFFFFFE;//32BIT INT, 64BIT OS NEED UPDATE
	myLeaderIndex=testerIndex&mask;
	myGroupSize=2;
	myTileCount=DepthBufferSnapShotNumber/2;
	if (myTesterIndex==myLeaderIndex)
	{
		myTileIndex=0;
		mySendIndex=myTileIndex+myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=true;
		myBuddyIndex=myLeaderIndex+1;
	}
	else
	{
		myTileIndex=DepthBufferSnapShotNumber/2;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=false;
		myBuddyIndex=myLeaderIndex;
	}
}
	
bool BinarySwapDescriptor::next()
{
	mask<<=1;
	if ((mask&TesterCount)==0)
	{
		return false;
	}
	myLeaderIndex=mask&myTesterIndex;
	myTileCount/=2;
	if (myTesterIndex<myLeaderIndex+myGroupSize)
	{
		myBuddyIndex=myTesterIndex+myGroupSize;
		myRecvIndex=myTileIndex;
		mySendIndex=myTileIndex+myTileCount;
	}
	else
	{
		myBuddyIndex=myTesterIndex-myGroupSize;
		myTileIndex+=myTileCount;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
	}
	myGroupSize*=2;
	return true;
}		


void BinarySwapDescriptor::display()
{
	
	printf("\n******************************\nBinarySwapDescriptor: myTesterIndex=%d, myLeaderIndex=%d,myGroupSize=%d,\
		\nmyBuddyIndex=%d,myTileIndex=%d,myTileCount=%d,\n\
		mySendIndex=%d,myRecvIndex=%d,mask=%X\n", myTesterIndex, myLeaderIndex, myGroupSize,myBuddyIndex,\ 
		myTileIndex, myTileCount, mySendIndex, myRecvIndex, mask);
}


void TesterContext::init()
{
	int i, j;
	MPI_Group commGroup;
	int testerRanks[TesterCount];

	myTesterIndex=mpiContext.myRank-TesterHostStartIndex;

	colorBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(colorBuffer);
	depthBufferCount=2;

	depthBuffer=new float*[depthBufferCount];
	CHECK_NEW(depthBuffer);

	for (i=0; i<depthBufferCount; i++)
	{
		depthBuffer[i]=new float[DepthBufferSize];
		CHECK_NEW(depthBuffer[i]);
	}	
	myDepthBuffer=depthBuffer[0];
	hisDepthBuffer=depthBuffer[1];
	//myRecvDepthBuffer=depthBuffer[TesterCount];
	//myDepthBuffer=depthBuffer[myTesterIndex];
	//the laster one

	//let's make an independent buffer
	snapShotBuffer=new float[TesterCount*DepthBufferSnapShotNumber*2];
	CHECK_NEW(snapShotBuffer);
	snapShots=new float*[TesterCount];
	CHECK_NEW(snapShots);

	for (i=0; i<TesterCount; i++)
	{
		//in order to use "mpi_allgather" we need to make it continuous memory. so
		//snapShots now is simply pointer arrays
		snapShots[i]=snapShotBuffer+DepthBufferSnapShotNumber*2*i;
		//don't forget the *2 means min and max
		//snapShots[i]=new float[DepthBufferSnapShotNumber*2];
		//CHECK_NEW(snapShots[i]);
	}

	visibleTable=new ubyte*[depthBufferCount];
	CHECK_NEW(visibleTable);
	for (i=0; i<depthBufferCount; i++)
	{
		visibleTable[i]=new ubyte[ColorBufferSize];
		CHECK_NEW(visibleTable[i]);
	}
	myVisibleTable=visibleTable[0];
	hisVisibleTable=visibleTable[1];
	
	/////////////////////////////
	//compressed visible table buffer
	compressedVisibleTable=new ubyte[VisibleTableSize];
	CHECK_NEW(compressedVisibleTable);
	///////////////////////////////	

	/////////////////////////////////////////////////////////////////////////////////////////
	//requests!!
	//half recv, half send
	snapShotSendingRequests=new MPI_Request[TesterCount];
	CHECK_NEW(snapShotSendingRequests);
	snapShotRecvingRequests=new MPI_Request[TesterCount];
	CHECK_NEW(snapShotRecvingRequests);
	for (i=0; i<TesterCount; i++)
	{
		snapShotSendingRequests[i]=snapShotRecvingRequests[i]=MPI_REQUEST_NULL;		
	}

	//because exchange now becomes one-way

	visibleTableExchangeRequests= new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(visibleTableExchangeRequests);
#ifdef 	USING_BINARY_SWAP
	binarySwapDescriptor=new BinarySwapDescriptor;
	CHECK_NEW(binarySwapDescriptor);
#endif


	depthBufferExchangeRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(depthBufferExchangeRequests);

	visibleTableRenderRequests=new MPI_Request*[RenderCount];//*RenderCount];	
	CHECK_NEW(visibleTableRenderRequests);

	for (i=0; i<RenderCount; i++)
	{
		visibleTableRenderRequests[i]=new MPI_Request[DepthBufferSnapShotNumber];
		CHECK_NEW(visibleTableRenderRequests[i]);
		for (j=0; j<DepthBufferSnapShotNumber; j++)
		{
			visibleTableRenderRequests[i][j]=MPI_REQUEST_NULL;
		}
	}

	visibleTableDisplayerRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(visibleTableDisplayerRequests);

	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableDisplayerRequests[i]=MPI_REQUEST_NULL;
	}

	exchangeDescriptor=new DepthBufferExchangeDescriptor[DepthBufferSnapShotNumber];
	CHECK_NEW(exchangeDescriptor);

	for (i=0; i<TesterCount; i++)
	{
		testerRanks[i]=TesterHostStartIndex+i;
	}

	SAFE_CALL(MPI_Comm_group(MPI_COMM_WORLD, &commGroup));
	SAFE_CALL(MPI_Group_incl(commGroup, TesterCount, testerRanks, &testerGroup));
	SAFE_CALL(MPI_Comm_create(MPI_COMM_WORLD, testerGroup, &testerComm));


	octreeLoader.initialize(true, true);
	renderEngine.init(&octreeLoader);
	renderEngine.lighting.lightEnabled=false;
///////////////////////////////////////////////////
	masterContext.init();
	masterContext.viewCollector.startCollecting(masterContext.renderEngine.drawingContext.center);
        masterContext.knapsack(masterContext.viewCollector.octreeNodeVector.size(), &(*masterContext.viewCollector.octreeNodeVector.begin()), false);
///////////////////////////////////////////////////

	renderIndexHashTable=new ubyte[Octree::infoStruct.nodeCount];
	CHECK_NEW(renderIndexHashTable);
	snapShotCandidate=new int[DepthBufferSnapShotNumber];
	CHECK_NEW(snapShotCandidate);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadBuffer(GL_BACK);

	createRenderIndexHashTable();
	int offset=0;
	for (i=0; i<myTesterIndex; i++)
	{
		offset+=mpiContext.renderContext->testerShare[i];
	}
	octreeLoader.addCandidates(mpiContext.renderContext->testerShare[myTesterIndex], 
		mpiContext.testerShareArray+offset);
	octreeLoader.swapBuffer();
	readColorBuffer();
	readDepthBuffer();
	processDepthBuffer();


//////////////////////////////////////////////////////////
#ifdef 	DEBUGGING
#ifdef LINUX_VERSION
	openlog("renderBox", 0,0);
#endif
#endif


}


//create a hashtable such that the octree node index is mapped to its render
void TesterContext::createRenderIndexHashTable()
{
	int offset=0;
	for (int i=0; i<RenderCount; i++)
	{
		for (int j=0; j<mpiContext.renderContext->renderShare[i]; j++)
		{
			//mapping from octree node index to render node index
			renderIndexHashTable[mpiContext.renderShareArray[offset+j]]=i;
		}
		offset+=mpiContext.renderContext->renderShare[i];
	}
}


void TesterContext::keyboardCallback(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 'e':
	case 'E':
	case 'n':
	case 'N':
		break;
	case 0:
		break;
	default:
		renderEngine.keyboardCallback(key, x, y);
		break;
	}
	glutPostRedisplay();
}


void TesterContext::displayCallback()
{
	DEBUG_INFO("tester begins displaycallback......\n");

	PROFILE_BEGIN(ALL);

#ifdef	USING_SNAPSHOT
	exchangeDepthBufferSnapShot();
#endif

	PROFILE_BEGIN(BT);
	exchangeDepthBuffer();
	PROFILE_END(BT);
	PROFILE_END(ALL);
	DEBUG_INFO("tester finishes displaycallback......\n");
}

void TesterContext::readColorBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, DepthBufferAtomWidth, 
				GL_RGBA, GL_UNSIGNED_BYTE, colorBuffer+4*(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize);
		}
	}
}

void TesterContext::readDepthBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, DepthBufferAtomWidth, 
				GL_DEPTH_COMPONENT, GL_FLOAT, 
				myDepthBuffer+(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize);
		}
	}
}

#ifdef USING_SNAPSHOT
void TesterContext::processDepthBuffer()
{
	int i, j, k, index, offset, pos;
	float* minPtr, *maxPtr, *curPtr;
	unsigned int* colorPtr;
	//this must be "little-indian"
	const unsigned int HeadMask=0x00FFFFFF;
	for (i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (j=0; j<DepthBufferSnapShotWidth; j++)
		{
			index=i*DepthBufferSnapShotWidth+j;
			minPtr=snapShots[myTesterIndex]+index*2;
			maxPtr=snapShots[myTesterIndex]+index*2+1;
			//initialize
			*minPtr=*maxPtr=myDepthBuffer[index*DepthBufferAtomSize];
			offset=index*DepthBufferAtomSize;
			for (k=0; k<DepthBufferAtomSize; k++)
			{
				curPtr=myDepthBuffer+ offset+k;
				if (*curPtr<*minPtr)
				{
					*minPtr=*curPtr;
				}
				if (*curPtr>*maxPtr)
				{
					*maxPtr=*curPtr;
				}
				//here we don't restrict ourselves as big-endian or little-endian
				//cause it is an integer written from "fakecolor.data"
				colorPtr=(unsigned int*)(colorBuffer+(offset+k)*4);
				//but here it is still not restricted as little-endian
				//definitely it is IMPORTANT! but I don't understand why now.
				pos=(*colorPtr)&HeadMask;//this must be from my VERY old version which uses "alpha" to store index
				//pos=*colorPtr;//I forget this purpose, but it should be useless
				
				if (*curPtr==1)
				{
					//invisible
					myVisibleTable[offset+k]=RenderCount;//invisible Color;
				}
				else
				{
					//visible
					myVisibleTable[offset+k]=renderIndexHashTable[pos];
					//assert(renderIndexHashTable[pos]<RenderCount);
				}
			}
		}
	}
}
#else
void TesterContext::processDepthBuffer()
{
	int i, j, k, index, offset, pos;
	unsigned int* colorPtr;
	float* curPtr;
	//this must be "little-indian"
	const unsigned int HeadMask=0x00FFFFFF;
	for (i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (j=0; j<DepthBufferSnapShotWidth; j++)
		{
			index=i*DepthBufferSnapShotWidth+j;

			offset=index*DepthBufferAtomSize;
			for (k=0; k<DepthBufferAtomSize; k++)
			{
				curPtr=myDepthBuffer+ offset+k;

				//here we don't restrict ourselves as big-endian or little-endian
				//cause it is an integer written from "fakecolor.data"
				colorPtr=(unsigned int*)(colorBuffer+(offset+k)*4);
				//but here it is still not restricted as little-endian
				//definitely it is IMPORTANT! but I don't understand why now.
				//alpha should be 0, then there is no need for this masking
				pos=(*colorPtr)&HeadMask;//this must be from my VERY old version which uses "alpha" to store index
				//pos=*colorPtr;//I forget this purpose, but it should be useless
				
				if (*curPtr==1)
				{
					//invisible
					myVisibleTable[offset+k]=RenderCount;//invisible Color;
				}
				else
				{
					//visible
					myVisibleTable[offset+k]=renderIndexHashTable[pos];
					//assert(renderIndexHashTable[pos]<RenderCount);
				}
			}
		}
	}
}
#endif


void TesterContext::exchangeDepthBufferSnapShot()
{
	int source;
	int i, j;
	
	//printf("Tester %d exchange snapshots begins\n", myTesterIndex);
#ifndef USING_ALL_GATHER
	for (i=0; i<TesterCount; i++)
	{
		if (i!=myTesterIndex)
		{		
			source=TesterHostStartIndex+i;
		
			SAFE_CALL(MPI_Irecv(snapShots[i], DepthBufferSnapShotNumber*2, MPI_FLOAT, source,
				DEPTH_BUFFER_SNAP_SHOT, MPI_COMM_WORLD, snapShotRecvingRequests+i));
			
	
			SAFE_CALL(MPI_Isend(snapShots[myTesterIndex], DepthBufferSnapShotNumber*2, MPI_FLOAT, source,
				DEPTH_BUFFER_SNAP_SHOT, MPI_COMM_WORLD, snapShotSendingRequests+i));
			
		}
	}
	SAFE_CALL(MPI_Waitall(TesterCount, snapShotSendingRequests, MPI_STATUSES_IGNORE));
	SAFE_CALL(MPI_Waitall(TesterCount, snapShotRecvingRequests, MPI_STATUSES_IGNORE));
#else
	SAFE_CALL(MPI_Allgather(snapShots[myTesterIndex], DepthBufferSnapShotNumber*2, MPI_FLOAT, snapShots[0], 
		DepthBufferSnapShotNumber*2, MPI_FLOAT, testerComm));


#endif
	//printf("Tester %d exchange snapshots finished\n", myTesterIndex);
}



#ifdef	USING_SNAPSHOT
//always compare with minPtr
void TesterContext::exchangeDepthBuffer()
{
	int i, j, tileCount=0, startingIndex, index;
	float* curPtr, *candPtr, *myPtr;
	
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		//choose a default for comparison
		snapShotCandidate[i]=myTesterIndex;
		candPtr=snapShots[myTesterIndex]+i*2;
		//depthBufferExchangeRequests[i]=MPI_REQUEST_NULL;
		//visibleTableExchangeRequests[i]=MPI_REQUEST_NULL;
		for (j=0; j<TesterCount; j++)
		{
			curPtr=snapShots[j]+i*2;
			//when equal, we choose by index
			if (*curPtr<*candPtr||*curPtr==*candPtr&&j<snapShotCandidate[i])
			{
				candPtr=curPtr;
				snapShotCandidate[i]=j;
			}
		}
		
		myPtr=snapShots[myTesterIndex]+i*2;
		candPtr=snapShots[snapShotCandidate[i]]+i*2;

		exchangeDescriptor[i].testerCount=0;
		exchangeDescriptor[i].myIndex=-1;
		depthBufferExchangeRequests[i]=MPI_REQUEST_NULL;
		visibleTableExchangeRequests[i]=MPI_REQUEST_NULL;

		startingIndex=i%TesterCount;
		//there are three cases: 1. index=candidate,2. index=modulo 3. my_small<candiate_big		
		//if (snapShotCandidate[i]!=myTesterIndex&&myPtr[0]<canPtr[1]||snapShotCandidate[i]==myTesterIndex)
		if (myPtr[0]<candPtr[1]||startingIndex==myTesterIndex)
		{
			exchangeDescriptor[i].testerCount=1;
			exchangeDescriptor[i].testers[0]=startingIndex;
			//because we skip the first in following code, we have to add here
			if (startingIndex==myTesterIndex)
			{
				exchangeDescriptor[i].myIndex=0;
			}
			//now we check how many Testers for this particular snapshot
			for (j=1; j<TesterCount; j++)
			{			
				//here we only want to scramble pairs a bit
				index=(startingIndex+j)%TesterCount;
				curPtr=snapShots[index]+i*2;
				//if its smallest is smaller than our biggest
				if (curPtr[0]<candPtr[1])
				{
					if (index==myTesterIndex)
					{
						exchangeDescriptor[i].myIndex=exchangeDescriptor[i].testerCount;
					}
					exchangeDescriptor[i].testers[exchangeDescriptor[i].testerCount]=index;
					exchangeDescriptor[i].testerCount++;
				}
			}
			if (exchangeDescriptor[i].testerCount>1)
			{
				startExchangeDepthBuffer(i);
				tileCount++;
			}
		}
		//printf("exchange descriptor no.%d\n", i);
		//exchangeDescriptor[i].display();
	}
	//if (myTesterIndex==0)
	//printf("before doExchangeDepthBuffer#%d\n", tileCount);
	doExchangeDepthBuffer(tileCount);
}
#else
//without snapshot
#ifdef	USING_BINARY_SWAP
void TesterContext::exchangeDepthBuffer()
{
	int length, i;
	int depthSize;
	int buddyIndex;
	float* depthSendPtr, *depthRecvPtr;
	float* myOwnDPtr;
	unsigned char* myVPtr, *hisVPtr;
	unsigned char*myOwnVPtr;
	MPI_Status status;
	unsigned char* visibleSendPtr, *visibleRecvPtr;
	if (TesterCount==1)
	{
		return;
	}
	binarySwapDescriptor->init(myTesterIndex);
	do
	{
		depthSize=binarySwapDescriptor->myTileCount*DepthBufferAtomSize;
		depthRecvPtr=hisDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		depthSendPtr=myDepthBuffer+binarySwapDescriptor->mySendIndex*DepthBufferAtomSize;
		visibleSendPtr=compressedVisibleTable+binarySwapDescriptor->mySendIndex*DepthBufferAtomSize*VisibleTableBufferRate;
		visibleRecvPtr=compressedVisibleTable+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize*VisibleTableBufferRate;		
		buddyIndex=TesterHostStartIndex+binarySwapDescriptor->myBuddyIndex;

		myVPtr=myVisibleTable+binarySwapDescriptor->mySendIndex*DepthBufferAtomSize;
		hisVPtr=hisVisibleTable+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		
		//the sequence maybe a bit optimized if you post request before compression		
		SAFE_CALL(MPI_Irecv(visibleRecvPtr, depthSize*VisibleTableBufferRate, MPI_UNSIGNED_CHAR, buddyIndex,
			BINARY_SWAP_VISIBLE_TABLE, MPI_COMM_WORLD, visibleTableExchangeRequests+binarySwapDescriptor->myBuddyIndex));


		SAFE_CALL(MPI_Irecv(depthRecvPtr, depthSize, MPI_FLOAT, buddyIndex,
			BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex));


		SAFE_CALL(MPI_Isend(depthSendPtr, depthSize, MPI_FLOAT, buddyIndex, BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, 
			depthBufferExchangeRequests+myTesterIndex));

		 		
		length=compressBuffer(myVPtr, depthSize, visibleSendPtr, depthSize*VisibleTableBufferRate);

		SAFE_CALL(MPI_Isend(visibleSendPtr, length, MPI_UNSIGNED_CHAR, buddyIndex, BINARY_SWAP_VISIBLE_TABLE, MPI_COMM_WORLD,
			visibleTableExchangeRequests+myTesterIndex));

		//a little optimization is to test compressed first for decompression
		SAFE_CALL(MPI_Wait(visibleTableExchangeRequests+binarySwapDescriptor->myBuddyIndex, &status));
		SAFE_CALL(MPI_Get_count(&status, MPI_UNSIGNED_CHAR, &length));
		decompressBuffer(visibleRecvPtr, length, hisVPtr, depthSize);

		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex, MPI_STATUS_IGNORE));
		
		//mySend is what I send, myRecv is what I recv, but the same index is used for buffer remains to compared with recved
		myOwnDPtr=myDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		myOwnVPtr=myVisibleTable+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;

		for (i=0; i<depthSize; i++)
		{
			if (depthRecvPtr[i]<myOwnDPtr[i])
			{
				myOwnVPtr[i]=hisVPtr[i];
				myOwnDPtr[i]=depthRecvPtr[i];
			}
		}
		//sending can always wait for testing
		SAFE_CALL(MPI_Wait(visibleTableExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));

		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));
		//debugging...	
		//memset(hisDepthBuffer, 0, DepthBufferSize*sizeof(float));
		//memset(hisVisibleTable, 0, VisibleTableSize);
		if (myTesterIndex==10)
		{			
			//binarySwapDescriptor->display();
		}
	}
	while (binarySwapDescriptor->next());
}

#else //without using binary_swap, purely binary-tree-like exchange
void TesterContext::exchangeDepthBuffer()
{
	int i, j, tileCount=0, startingIndex, index;
	float* curPtr, *candPtr, *myPtr;
	if (TesterCount==1)
	{
		return;
	}	
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		depthBufferExchangeRequests[i]=MPI_REQUEST_NULL;
		visibleTableExchangeRequests[i]=MPI_REQUEST_NULL;

		startingIndex=i%TesterCount;
		exchangeDescriptor[i].testerCount=0;
		//because we skip the first in following code, we have to add here
		//now we check how many Testers for this particular snapshot
		for (j=0; j<TesterCount; j++)
		{			
			//here we only want to scramble pairs a bit
			index=(startingIndex+j)%TesterCount;
			if (index==myTesterIndex)
			{
				exchangeDescriptor[i].myIndex=exchangeDescriptor[i].testerCount;
			}
			exchangeDescriptor[i].testers[exchangeDescriptor[i].testerCount]=index;
			exchangeDescriptor[i].testerCount++;
		}
		startExchangeDepthBuffer(i);

		if (exchangeDescriptor[i].testerCount>1)
		{
			tileCount++;
		}
		//printf("exchange descriptor no.%d\n", i);
		//exchangeDescriptor[i].display();
	}
	//if (myTesterIndex==0)
	//printf("before doExchangeDepthBuffer#%d\n", tileCount);
	doExchangeDepthBuffer(tileCount);
}
#endif//binary_swap
#endif



void TesterContext::startExchangeDepthBuffer(int tileIndex)
{	
//	if (myTesterIndex==0)
//	printf("START EXCHANGE DEPTH BUFFER#%d\n", tileIndex);
	if (exchangeDescriptor[tileIndex].testerCount==1)
	{
		return;
	}

	if (exchangeDescriptor[tileIndex].myIndex%2==1)
	{
		sendExchangeDepthBuffer(tileIndex);
	}
	else
	{
		if (exchangeDescriptor[tileIndex].myIndex+1!=exchangeDescriptor[tileIndex].testerCount)
		{
			//receiving
			recvExchangeDepthBuffer(tileIndex);
		}
		else
		{
			//directly we enter next round
			enterNextRound(tileIndex);
			startExchangeDepthBuffer(tileIndex);//recursion
		}			
	}
}

void TesterContext::enterNextRound(int tileIndex)
{
	int i;
	for (i=0; i*2<exchangeDescriptor[tileIndex].testerCount; i++)
	{
		exchangeDescriptor[tileIndex].testers[i]=exchangeDescriptor[tileIndex].testers[i*2];
		
	}
	exchangeDescriptor[tileIndex].testerCount=i;
	exchangeDescriptor[tileIndex].myIndex/=2;
}

void TesterContext::finishExchangeDepthBuffer(int tileIndex)
{
	int i;
	ubyte*myTable, *hisTable;	
	float*myPtr, *hisPtr;

	myTable=myVisibleTable+tileIndex*DepthBufferAtomSize;
	hisTable=hisVisibleTable+tileIndex*DepthBufferAtomSize;
	myPtr=myDepthBuffer+tileIndex*DepthBufferAtomSize;
	hisPtr=hisDepthBuffer+tileIndex*DepthBufferAtomSize;

	//merging	
	for (i=0; i<DepthBufferAtomSize; i++)
	{
		if (hisPtr[i]<myPtr[i])
		{
			myPtr[i]=hisPtr[i];
			myTable[i]=hisTable[i];
		}
	}
}

void TesterContext::sendExchangeDepthBuffer(int tileIndex)
{
	int length, targetTester, targetIndex;
	//sending and there is always a target
	//assert(exchangeDescriptor[tileIndex].myIndex>=0);
	targetIndex=exchangeDescriptor[tileIndex].myIndex-1;
	//assert(targetIndex>=0);

	targetTester=exchangeDescriptor[tileIndex].testers[targetIndex];

	//printf("tileIndex#%d:tester#%d: send to %d\n", tileIndex, myTesterIndex, targetTester);	


	SAFE_CALL(MPI_Isend(myDepthBuffer+tileIndex*DepthBufferAtomSize, DepthBufferAtomSize, MPI_FLOAT, 
		targetTester+TesterHostStartIndex, tileIndex, MPI_COMM_WORLD, 
		depthBufferExchangeRequests+tileIndex));
	length=compressBuffer(myVisibleTable+tileIndex*DepthBufferAtomSize, DepthBufferAtomSize,
		compressedVisibleTable+tileIndex*DepthBufferAtomSize*VisibleTableBufferRate, 
		DepthBufferAtomSize*VisibleTableBufferRate);

	SAFE_CALL(MPI_Isend(compressedVisibleTable+ tileIndex*DepthBufferAtomSize*VisibleTableBufferRate, 
		length,	MPI_UNSIGNED_CHAR, targetTester+TesterHostStartIndex, tileIndex+DepthBufferSnapShotNumber, 
		MPI_COMM_WORLD,	visibleTableExchangeRequests+tileIndex));
}


void TesterContext::recvExchangeDepthBuffer(int tileIndex)
{
	int targetTester, targetIndex;
	//if (myTesterIndex==0)
	//printf("recv exchange tile#%d\n", tileIndex);
	//assert(exchangeDescriptor[tileIndex].myIndex>=0);

	targetIndex=exchangeDescriptor[tileIndex].myIndex+1;

	//assert(targetIndex<exchangeDescriptor[tileIndex].testerCount);

	targetTester=exchangeDescriptor[tileIndex].testers[targetIndex];

	//printf("tileIndex#%d:tester#%d: recv from %d\n", tileIndex, myTesterIndex, targetTester);	

	SAFE_CALL(MPI_Irecv(hisDepthBuffer+ tileIndex*DepthBufferAtomSize, DepthBufferAtomSize, MPI_FLOAT, 
		targetTester+TesterHostStartIndex, tileIndex, MPI_COMM_WORLD, 
		depthBufferExchangeRequests+tileIndex));
	SAFE_CALL(MPI_Irecv(compressedVisibleTable+ tileIndex*DepthBufferAtomSize*VisibleTableBufferRate, 
		DepthBufferAtomSize*VisibleTableBufferRate, MPI_UNSIGNED_CHAR, targetTester+TesterHostStartIndex, 
		tileIndex+DepthBufferSnapShotNumber, MPI_COMM_WORLD, 
		visibleTableExchangeRequests+tileIndex));
}



void TesterContext::doExchangeDepthBuffer(int tileCount)
{	
	int length, i;
	int tileIndex=-1;
	MPI_Status status;
	
	while (tileCount>0)
	{
		SAFE_CALL(MPI_Waitany(DepthBufferSnapShotNumber, depthBufferExchangeRequests, &tileIndex, MPI_STATUS_IGNORE));
		assert(tileIndex>=0&&tileIndex<DepthBufferSnapShotNumber);

		SAFE_CALL(MPI_Wait(visibleTableExchangeRequests+tileIndex, &status));
	

		if (exchangeDescriptor[tileIndex].myIndex%2==0)
		{
			SAFE_CALL(MPI_Get_count(&status, MPI_UNSIGNED_CHAR, &length));
		
			length=decompressBuffer(compressedVisibleTable+ tileIndex*DepthBufferAtomSize*VisibleTableBufferRate, length,
				hisVisibleTable+tileIndex*DepthBufferAtomSize, DepthBufferAtomSize);
			finishExchangeDepthBuffer(tileIndex);
			enterNextRound(tileIndex);
			
			startExchangeDepthBuffer(tileIndex);
			if (exchangeDescriptor[tileIndex].testerCount==1)
			{
				tileCount--;
			}
		}
		else
		{
			tileCount--;
		}
	}
}

#ifdef	USING_BINARY_SWAP
//assuming TesterCount, TileCount are all power of 2
void TesterContext::sendVisiblePixelInfo()
{
	int i, j, tileIndex;
	ubyte* visibleTablePtr, *visibleSendPtr;
	int visibleTableSendingSize;
	int displayerIndex;
#ifdef	USING_SEPARATE_DISPLAYER
	displayerIndex=DisplayerIndex;
#else
	displayerIndex=MasterIndex;
#endif

	tileIndex=binarySwapDescriptor->myRecvIndex;
	assert(tileIndex>=0&&tileIndex<DepthBufferSnapShotNumber);

	visibleTablePtr=myVisibleTable+tileIndex*DepthBufferAtomSize;

	visibleSendPtr=compressedVisibleTable+tileIndex*DepthBufferAtomSize*VisibleTableBufferRate;

	//assert(binarySwapDescriptor->myTileCount==1);
	for (i=0; i<binarySwapDescriptor->myTileCount; i++)
	{		
		visibleTableSendingSize=compressBuffer(visibleTablePtr+i*DepthBufferAtomSize, DepthBufferAtomSize, 
			visibleSendPtr+i*DepthBufferAtomSize*VisibleTableBufferRate, DepthBufferAtomSize*VisibleTableBufferRate);
		SAFE_CALL(MPI_Isend(visibleSendPtr+i*DepthBufferAtomSize*VisibleTableBufferRate, visibleTableSendingSize,
			MPI_UNSIGNED_CHAR, displayerIndex, tileIndex+i, MPI_COMM_WORLD, visibleTableDisplayerRequests+tileIndex+i));
		for (j=0; j<RenderCount; j++)
		{
			SAFE_CALL(MPI_Isend(visibleSendPtr+i*DepthBufferAtomSize*VisibleTableBufferRate, visibleTableSendingSize, MPI_UNSIGNED_CHAR, 
				RenderHostStartIndex+j, tileIndex+i, MPI_COMM_WORLD, 
				visibleTableRenderRequests[j]+tileIndex+i));
		}
	}
	//printf("Tester%d sends visible pixelInfo of total %d\n", myTesterIndex, counter);
}

#else
//since we put module at index 0 of exchange descriptor tester array, it will always converge
void TesterContext::sendVisiblePixelInfo()
{
	int i, j;
	//int counter=0;
	//for each display tile
	ubyte* visibleTablePtr;
	int displayerIndex;
	int visibleTableSendingSize;
#ifdef	USING_SEPARATE_DISPLAYER
	displayerIndex=DisplayerIndex;
#else
	displayerIndex=MasterIndex;
#endif

	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		//if current tester is responsible for visiblePixelInfo sending...
		if (i%TesterCount==myTesterIndex)
		{
			//counter++;
			visibleTableDisplayerRequests[i]=MPI_REQUEST_NULL;
			//send to displayer
			//still we need the pointer for parameter	
			visibleTablePtr=myVisibleTable+i*DepthBufferAtomSize;

			visibleTableSendingSize=compressBuffer(visibleTablePtr, DepthBufferAtomSize, 
				compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, 
				DepthBufferAtomSize*VisibleTableBufferRate);
			//assert(visibleTableSendingSize!=-1);
			
			visibleTablePtr=compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate;

			SAFE_CALL(MPI_Isend(visibleTablePtr, visibleTableSendingSize,
				MPI_UNSIGNED_CHAR, displayerIndex, i, MPI_COMM_WORLD, visibleTableDisplayerRequests+i));
			for (j=0; j<RenderCount; j++)
			{
				visibleTableRenderRequests[j][i]=MPI_REQUEST_NULL;
				//send to each renderer, with tag of display tile index
				SAFE_CALL(MPI_Isend(visibleTablePtr, visibleTableSendingSize,
					MPI_UNSIGNED_CHAR, RenderHostStartIndex+j, i, MPI_COMM_WORLD, 
				visibleTableRenderRequests[j]+i));
			}
		}
	}
	//printf("Tester%d sends visible pixelInfo of total %d\n", myTesterIndex, counter);
}
#endif//binary_swap


void TesterContext::testVisiblePixelInfoSending()
{
	int i;
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableDisplayerRequests, MPI_STATUSES_IGNORE));

	for (i=0; i<RenderCount; i++)
	{
		SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRenderRequests[i], MPI_STATUSES_IGNORE));		
	}
	//printf("Tester%d finishes sending visible pixelInfo\n", myTesterIndex);
}





//#define TESTING 1





unsigned long frameCounter=0;
unsigned long lastFrameNo=0;
unsigned long lastTick=0;
unsigned long currentTick=0;

void displayCallbackProc();

void keyboardCallbackProc(unsigned char key, int x, int y);

void idleCallbackProc();

char dumpFileName[32];

FILE* dumpPtr;

void dump()
{
	//PROFILE_DISPLAY();
	PROFILE_PRINT();
	if (dumpPtr!=NULL)
	{
		fclose(dumpPtr);
	}
}


int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);

	mpiContext.init();
	gethostname(mpiContext.nodeName, 32);
	printf("host %s starts at rank %d\n", mpiContext.nodeName, mpiContext.myRank);	

	glutInit(&argc, argv);
	getLastError();
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
	glutInitWindowSize(ScreenSize, ScreenSize);
#ifdef TESTING
	glutInitWindowPosition(5+(mpiContext.myRank%4)*ScreenSize, 5+(mpiContext.myRank%4)*ScreenSize);
#else
	glutInitWindowPosition(0, 0);
#endif
	strcpy(dumpFileName, mpiContext.nodeName);
	strcat(dumpFileName, ".txt");
	dumpPtr=fopen(dumpFileName, "w");
	if (dumpPtr==NULL)
	{
		printf("open dump file failed\n");
		exit(1);
	}
	mpiContext.profiler.filePtr=dumpPtr;
	if (atexit(dump)!=0)
	{
		printf("atexit fails\n");
		exit(10);
	}

	glutCreateWindow(testerContext.windowTitle);		
	testerContext.init();
	printf("tester rank#%d finishes init\n", mpiContext.myRank);

	glutDisplayFunc(displayCallbackProc);
	glutKeyboardFunc(keyboardCallbackProc);
	glutPostRedisplay();
	glutMainLoop();
	return 0;
}


void displayCallbackProc()
{
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	if (frameCounter!=0)
	{
		printf("frame#%u:fps=%f\n", frameCounter, 1000.0/(float)(currentTick-lastTick));
	}
	lastTick=currentTick;
	testerContext.displayCallback();
	frameCounter++;
	if (frameCounter==QUIT_FRAME_NUMBER)
	{
		mpiContext.quit();
	}
	glutSwapBuffers();
#ifdef USING_FORCE_REFRESH
	glutPostRedisplay();
#endif
}



void keyboardCallbackProc(unsigned char key, int x, int y)
{
	glutPostRedisplay();
}


void idleCallbackProc()
{

}


			
			












