/******************************************************************
author: qingzhe huang
date: Jul. 31, 2007
1. This module is a combination of both "Master" and "Displayer". Originally it
should be separated, at least logically. However, in order to run in more nodes, I 
have to save displayer to combine it with master nodes as displayer is not a heavy
job.
2. The most difficult part is "tile" which troubles me for many days. Originally I
don't really grasp the meaning of "glReadPixel" which is so powerful that allow programer
to read only "tiles". You see, in both Testers and Renders they are using "tiles" 
to divide screen. However, when sending pixel data, Renders send pixel data according
to "tiles". Then using "tile" as basic unit, you don't have to worry about "sequence".
i.e. all "linear" sequence in both "pixel-visible-info" and "pixel-data" are based on 
"tiles". You just move your pointer and finally write back to color buffer in "tiles".
The big advantage is that in network transmission data buffer needs to be consecutive and
you don't have to make extra copy if you read and write color buffer in "tiles". This 
may not be a problem at all for somebody. But for me at beginning I read whole color 
buffer in one shot and manually divide "tiles" by index which is really painful experience.
3. The so-called knapsack is a greedy for NP-complete problem. I don't want to waste 
too much time on it, so make it a simple greedy.


******************************************************************/


#include "config.h"
#include "master.h"
#include "dataStructure.h"
#include "viewCollector.h"
#include <vector>
#include <assert.h>
#include <GL/glut.h>

#ifndef  LINUX_VERSION
#include "except.h"
#endif

using namespace std;



char* MasterContext::windowTitle="Master Node";
extern MPIContext mpiContext;

int myIntComp(const void* firstInt, const void* secondInt)
{
	return ViewCollector::root[*(int*)(firstInt)].fNumber+
		ViewCollector::root[*(int*)(firstInt)].fRepeatNumber - 
		ViewCollector::root[*(int*)(secondInt)].fNumber-
		ViewCollector::root[*(int*)(secondInt)].fRepeatNumber;
}


void MasterContext::knapsack(int count, int* array, bool doSimple)
{
	int i;
	mpiContext.renderContext->indexCount=count;
	assert(count<MaxPossibleNodeCollectable);
	mpiContext.renderShareArray=(int*)(mpiContext.renderContextBuffer+sizeof(RenderContext));
	mpiContext.testerShareArray=(int*)(mpiContext.renderContextBuffer+sizeof(RenderContext)+sizeof(int)*count);

	if (!doSimple)
	{
		//this is for render
	
			//doKnapsack(count, array, RenderCount, mpiContext.renderContext->renderShare, renderShareCount, 
			//	renderVector, mpiContext.renderShareArray);
			doKnapsack(count, array, RENDER);

			doKnapsack(count, array, TESTER);
			//this is for tester
			//doKnapsack(count, array, TesterCount, mpiContext.renderContext->testerShare, testerShareCount,
			//	testerVector, mpiContext.testerShareArray);


	}
	else
	{
		memcpy(mpiContext.renderShareArray, array, sizeof(int)*count);
		memcpy(mpiContext.testerShareArray, array, sizeof(int)*count);
		for (i=0; i<RenderCount; i++)
		{
			if (i!=RenderCount-1)
			{
				mpiContext.renderContext->renderShare[i]=count/RenderCount;
			}
			else
			{
				mpiContext.renderContext->renderShare[i]=count-(count/RenderCount)*(RenderCount-1);
			}
		}
		for (i=0; i<TesterCount; i++)
		{
			if (i!=TesterCount-1)
			{
				mpiContext.renderContext->testerShare[i]=count/TesterCount;
			}
			else
			{
				mpiContext.renderContext->testerShare[i]=count-(count/TesterCount)*(TesterCount-1);
			}
		}
	}
	mpiContext.renderContextSize=sizeof(RenderContext)+count*2*sizeof(int);

}


void MasterContext::doKnapsack(int count, int* array, NodeStatus nodeStatus)
{
	int partyNumber;
	int* shareArray;
	int* shareCount;
	vector<int>* shareVector;
	int* output;

	ubyte* hashTable;
	int average;
	int i, j, index;
	int offset=1;

	
	//int myIntComp(const void* firstInt, const void* secondInt);
	switch (nodeStatus)
	{
	case TESTER:
		partyNumber=TesterCount;
		shareArray=mpiContext.renderContext->testerShare;
		shareCount=testerShareCount;
		shareVector=testerVector;
		output=mpiContext.testerShareArray;
		hashTable=testerNodeHashTable;
		break;
	case RENDER:
		partyNumber=RenderCount;
		shareArray=mpiContext.renderContext->renderShare;
		shareCount=renderShareCount;
		shareVector=renderVector;
		output=mpiContext.renderShareArray;
		hashTable=renderNodeHashTable;
		break;
	default:
		printf("error!\n");
		exit(123);
	}
	if (partyNumber<1)
	{
		return;
	}

	for (i=0; i<partyNumber; i++)
	{
		shareArray[i]=0;
		shareCount[i]=0;
		shareVector[i].clear();
	}
	average=ViewCollector::faceCounter/partyNumber;
	//qsort(myVisibleCount, VisibleOctreeIndexCombinationLength, sizeof(int), myIntComp);
	//qsort(array, count, sizeof(int), myIntComp);

	j=0;		
	//let's evenly distribute the index set into "RenderHostCount" sets
	i=count-1;
	while (i>=0)
	{

#ifdef	USING_COHERENT_NODE
		if (hashTable[array[i]]!=255)//the owner take priority
		{
			index=hashTable[array[i]];
			if (shareCount[index]<average)
			{
				shareVector[index].push_back(array[i]);
				shareCount[index]+=ViewCollector::root[array[i]].fNumber+
					ViewCollector::root[array[i]].fRepeatNumber;
				i--;
			}
			else
			{
				//we reset its owner to nobody for next chance
				hashTable[array[i]]=255;
			}
		}
		else
		{
			if (shareCount[j]<average)
			{
				shareVector[j].push_back(array[i]);
				shareCount[j]+=ViewCollector::root[array[i]].fNumber+
					ViewCollector::root[array[i]].fRepeatNumber;
				hashTable[array[i]]=j;
				i--;
			}
		}
#else
		if (shareCount[j]<average)
		{
			shareVector[j].push_back(array[i]);
			shareCount[j]+=ViewCollector::root[array[i]].fNumber+
				ViewCollector::root[array[i]].fRepeatNumber;
			//hashTable[array[i]]=j;
			i--;
		}
#endif
			//going up
		j+=offset;
		if (j==partyNumber)
		{
			j=partyNumber-1;
			offset=-1;
		}
		//going down
		if (j==-1)
		{
			j=0;
			offset=1;
		}	
	}
	offset=0;

	if (mpiContext.renderContext->frameNumber==START_FRAME_NUMBER)
	{
		printf("\n[%s]share distribution\n", NodeName[nodeStatus]);
	}
	for (i=0; i<partyNumber; i++)
	{

		memcpy(output+offset, &(*shareVector[i].begin()), sizeof(int)*shareVector[i].size());
		shareArray[i]=shareVector[i].size();
		offset+=shareArray[i];
		if (mpiContext.renderContext->frameNumber==START_FRAME_NUMBER)
		{
			printf("[%d][%.2f\%]", shareCount[i], shareCount[i]*100.0/(float)ViewCollector::faceCounter);
		}
	}

}


void MasterContext::doKnapsack(int count, int* array, int partyNumber, int* shareArray, int* shareCount, 
				vector<int>* shareVector, int* output)
{
	int average;
	int i, j;
	int offset=1;

	int myIntComp(const void* firstInt, const void* secondInt);

	for (i=0; i<partyNumber; i++)
	{
		shareArray[i]=0;
		shareCount[i]=0;
		shareVector[i].clear();
	}
	average=ViewCollector::faceCounter/partyNumber;
	//qsort(myVisibleCount, VisibleOctreeIndexCombinationLength, sizeof(int), myIntComp);
	//qsort(array, count, sizeof(int), myIntComp);

	j=0;		
	//let's evenly distribute the index set into "RenderHostCount" sets
	i=count-1;
	while (i>=0)
	{
		if (shareCount[j]<average)
		{
			shareVector[j].push_back(array[i]);
			shareCount[j]+=ViewCollector::root[array[i]].fNumber+
				ViewCollector::root[array[i]].fRepeatNumber;
			i--;
		}
			//going up
		j+=offset;
		if (j==partyNumber)
		{
			j=partyNumber-1;
			offset=-1;
		}
		//going down
		if (j==-1)
		{
			j=0;
			offset=1;
		}	
	}
	offset=0;
	for (i=0; i<partyNumber; i++)
	{
		memcpy(output+offset, &(*shareVector[i].begin()), sizeof(int)*shareVector[i].size());
		shareArray[i]=shareVector[i].size();
		offset+=shareArray[i];
	}
}

void MasterContext::init()
{
	int i;
	octreeLoader.initialize(true);

	renderEngine.init(&octreeLoader);
	//printf("now renderEngine.movement\n");
	//renderEngine.movement.display();
/////////////////////////////////////////
	//renderEngine.movement.x-=octreeLoader.root->width/10.0;
	//renderEngine.movement.z+=octreeLoader.root->width/10.0;
	//renderEngine.updatePosition();
	


////////////////////////////

	viewCollector.initialize(octreeLoader.root);

	colorBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(colorBuffer);

	pixelDataBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(pixelDataBuffer);

	visibleTable=new ubyte[ColorBufferSize];
	CHECK_NEW(visibleTable);

	compressedVisibleTable=new ubyte[ColorBufferSize*VisibleTableBufferRate];
	CHECK_NEW(compressedVisibleTable);

	visibleTableRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(visibleTableRequests);
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
	}

	pixelDataRequests=new MPI_Request[RenderCount];
	CHECK_NEW(pixelDataRequests);
	for (i=0; i<RenderCount; i++)
	{
		pixelDataRequests[i]=MPI_REQUEST_NULL;
	}

	colorBufferRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(colorBufferRequests);
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		colorBufferRequests[i]=MPI_REQUEST_NULL;
	}
	renderVector=new vector<int>[RenderCount];
	CHECK_NEW(renderVector);
	testerVector=new vector<int>[TesterCount];
	CHECK_NEW(testerVector);
	renderShareCount=new int[RenderCount];
	CHECK_NEW(renderShareCount);
	testerShareCount=new int[TesterCount];
	CHECK_NEW(testerShareCount);

	naiveVisibleTableRequest=MPI_REQUEST_NULL;


	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glDrawBuffer(GL_BACK);
	//printf("leafnodenumber=%u\n", Octree::infoStruct.nodeCount);
	testerNodeHashTable=new ubyte[Octree::infoStruct.nodeCount];
	renderNodeHashTable=new ubyte[Octree::infoStruct.nodeCount];
	CHECK_NEW(testerNodeHashTable);
	CHECK_NEW(renderNodeHashTable);
	memset(testerNodeHashTable, 255, Octree::infoStruct.nodeCount);
	memset(renderNodeHashTable, 255, Octree::infoStruct.nodeCount);
	//printf("before knapsack\n");
	viewCollector.startCollecting(renderEngine.drawingContext.center);
	knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);
	printf("****************octree leaf node#[%d]****************\n", viewCollector.octreeNodeVector.size());
	mpiContext.sendRenderContext();
}


void MasterContext::uninit()
{
	delete[]testerShareCount;
	delete[]renderShareCount;
	delete[]testerVector;
	delete[]renderVector;
	delete[]pixelDataRequests;
	delete[]visibleTableRequests;
	delete[]visibleTable;
	delete[]pixelDataBuffer;
	delete[]colorBuffer;
	octreeLoader.uninitialize();
}


void MasterContext::recvVisiblePixelInfo()
{
	for (int i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
#ifndef USING_COMPRESSION
		SAFE_CALL(MPI_Irecv(visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize, MPI_UNSIGNED_CHAR,
			MPI_ANY_SOURCE, i, MPI_COMM_WORLD, visibleTableRequests+i));
#else
		SAFE_CALL(MPI_Irecv(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, DepthBufferAtomSize*VisibleTableBufferRate,
			 MPI_UNSIGNED_CHAR, MPI_ANY_SOURCE, i, MPI_COMM_WORLD, visibleTableRequests+i));
#endif		
	}
	DEBUG_INFO("master receives visible pixel info \n");
	//	runningStruct[RenderHostCount].arrayIndex, renderContext.frameNumber);
}

//let us use wait_any later for optimization
void MasterContext::testVisiblePixelInfoRecving()
{
	int i;
#ifdef 	USING_COMPRESSION
	MPI_Status status[DepthBufferSnapShotNumber];
	int length;
	
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, status));
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		SAFE_CALL(MPI_Get_count(status+i, MPI_UNSIGNED_CHAR, &length));
		length=decompressBuffer(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, length, 
			visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize);
		assert(length==DepthBufferAtomSize);
	}
#else
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, MPI_STATUSES_IGNORE));
#endif
	DEBUG_INFO("master finishes visible pixel info receiving\n");
}

void MasterContext::displayerPrepareRecvPixelData()
{
	int i;
	for (i=0; i<RenderCount+1; i++)
	{
		renderPixelCount[i]=0;
	}
	//counting how many pixel from each render, this is purely to save memory
	//so that in total we only need one color buffer size of buffer to hold all pixel data
	//from all renders. 
	for (i=0; i<ColorBufferSize; i++)
	{
		renderPixelCount[visibleTable[i]]++;
	}
}


void MasterContext::recvPixelData()
{
	int i, offset=0;	
	for (i=0; i<RenderCount; i++)
	{
		//here we use single size of color buffer to hold all pixel data from all renders
		renderPixelPtr[i]=pixelDataBuffer+offset;
		if (renderPixelCount[i]>0)
		{
			SAFE_CALL(MPI_Irecv(renderPixelPtr[i], renderPixelCount[i]*4, MPI_UNSIGNED_CHAR, 
				RenderHostStartIndex+i, PIXEL_DATA, MPI_COMM_WORLD, pixelDataRequests+i));
			DEBUG_INFO("master begin recv pixel data\n");
		}
		offset+=renderPixelCount[i]*4;		
		//printf("displayer expects to receives pixel data of %d from render#%d \n", renderPixelCount[i], i);
	}
	DEBUG_INFO("master receives pixel data\n");
}

//first let's do the simplest
void MasterContext::testPixelDataRecving()
{
	int i;
	SAFE_CALL(MPI_Waitall(RenderCount, pixelDataRequests, MPI_STATUSES_IGNORE));
	for (i=0; i<ColorBufferSize; i++)
	{
		if (visibleTable[i]<RenderCount)
		{
			//visible
			memcpy(colorBuffer+i*4, renderPixelPtr[visibleTable[i]], 4);
			renderPixelPtr[visibleTable[i]]+=4;
		}
		else
		{
			//set unvisible background to white
			memset(colorBuffer+i*4, 255, 4);
		}
	}
	//printf("displayer finishes receiving pixel data of frame#%d\n", mpiContext.renderContext->frameNumber);
}

#ifdef	USING_TESTER_RENDER
#ifdef	 USING_SEPARATE_DISPLAYER
void MasterContext::displayCallback()
{
	DEBUG_INFO("master begin displaycallback......\n");

	//PROFILE_PRINT();
	PROFILE_BEGIN(ALL);

	PROFILE_BEGIN(BMTRD);
	//mpiContext.sendRenderContext();
	mpiContext.testRenderContextSending();
	PROFILE_END(BMTRD);

	PROFILE_BEGIN(CM);
	viewCollector.startCollecting(renderEngine.drawingContext.center);
	knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);
	PROFILE_END(CM);

	mpiContext.sendRenderContext();

	PROFILE_END(ALL);
	DEBUG_INFO("master finishes displaycallback......\n");
#ifdef	DISPLAY_INFO
	if (mpiContext.renderContext->frameNumber%30==0)
	{
		displayInfo();
	}
#endif

}
#else
void MasterContext::displayCallback()
{
	DEBUG_INFO("master begin displaycallback\n");
	PROFILE_BEGIN(ALL);

	mpiContext.testRenderContextSending();
	if (!mpiContext.startMPI)
	{
		PROFILE_BEGIN(BMTRD);
		mpiContext.sendRenderContext();
		PROFILE_END(BMTRD);
	}
	else
	{
		DEBUG_INFO("master starts\n");
		PROFILE_BEGIN(CM);
		viewCollector.startCollecting(renderEngine.drawingContext.center);
		knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);
		PROFILE_END(CM);
		mpiContext.sendRenderContext();

		PROFILE_BEGIN(BRD);

		recvColorBuffer();
		testColorBufferRecving();
		PROFILE_END(BRD);
		PROFILE_BEGIN(CD);

		outputColorBuffer();
		PROFILE_END(CD);

		DEBUG_INFO("master finishes\n");
	}
	PROFILE_END(ALL);

}
#endif

#else//not tester_render
#ifdef	 USING_SEPARATE_DISPLAYER
void MasterContext::displayCallback()
{
	DEBUG_INFO("master begin displaycallback......\n");

	//PROFILE_PRINT();
	PROFILE_BEGIN(ALL);

	PROFILE_BEGIN(BMTRD);
	mpiContext.testRenderContextSending();
	PROFILE_END(BMTRD);

	PROFILE_BEGIN(CM);
	viewCollector.startCollecting(renderEngine.drawingContext.center);
	knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);
	PROFILE_END(CM);
	mpiContext.sendRenderContext();

	PROFILE_END(ALL);
	DEBUG_INFO("master finishes displaycallback......\n");
#ifdef	DISPLAY_INFO
	if (mpiContext.renderContext->frameNumber%30==0)
	{
		displayInfo();
	}
#endif

}

#else
void MasterContext::displayCallback()
{
	static bool firstTime=true;
	DEBUG_INFO("master begin displaycallback\n");
	PROFILE_BEGIN(ALL);
	mpiContext.testRenderContextSending();

	if (!mpiContext.startMPI)
	{
		mpiContext.sendRenderContext();
	}
	else
	{
	
		DEBUG_INFO("master starts\n");
		PROFILE_BEGIN(CM);
		if (!firstTime)
		{
			PROFILE_BEGIN(CD);
			outputColorBuffer();
			PROFILE_END(CD);
		}
		viewCollector.startCollecting(renderEngine.drawingContext.center);
		knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);
		mpiContext.sendRenderContext();
		PROFILE_END(CM);
		PROFILE_BEGIN(BTRD);

		recvVisiblePixelInfo();
		testVisiblePixelInfoRecving();
		PROFILE_END(BTRD);
		PROFILE_BEGIN(BRD);
		
		displayerPrepareRecvPixelData();
		recvPixelData();
		testPixelDataRecving();
		PROFILE_END(BRD);
		if (firstTime)
		{
			firstTime=false;
		}

		DEBUG_INFO("master finishes\n");
	}

	PROFILE_END(ALL);
}
#endif
#endif//tester_render


void MasterContext::outputColorBuffer()
{
	int i, j;
	//save context and setup to write color buffer
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, ScreenSize, 0, ScreenSize, 0, 100);
	//glOrtho(-ScreenSize/2, ScreenSize/2, -ScreenSize/2, ScreenSize/2, DefaultNearPlane, 100);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	//glRasterPos2f(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth);
	glRasterPos2f(0, 0);
	for (i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (j=0; j<DepthBufferSnapShotWidth; j++)
		{			
			//if (j%2==0)
			glRasterPos2f(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth);
			glDrawPixels(DepthBufferAtomWidth, DepthBufferAtomWidth, 
				GL_RGBA, GL_UNSIGNED_BYTE, 
				colorBuffer+(i*DepthBufferSnapShotWidth+j)*ColorBufferAtomSize);
		}
	}
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

void MasterContext::displayInfo()
{
	printf("\n****************running info***************************************\n");
	renderEngine.movement.display();
	printf("movement offset=%f\n", renderEngine.movementOffset);
	printf("octree leaf node number=%d\n", viewCollector.octreeNodeVector.size()); 
	printf("vertexRendered=%d, faceRendered=%d\n", viewCollector.vertexCounter, 
		viewCollector.faceCounter);
	printf("nearplane=%f, farplane=%f\n", renderEngine.drawingContext.zNear, renderEngine.drawingContext.zFar);
	printf("\n****************data info*****************************************\n");
	octreeLoader.root->infoStruct.display();
	printf("\n******************************************************************\n");
}


void MasterContext::keyboardCallback(unsigned char key, int x, int y)
{
	if (mpiContext.pendingKey!=0)
	{
		glutPostRedisplay();
		return;
	}
	else
	{
		mpiContext.pendingKey=key;
	}
	if (key==27)
	{
		mpiContext.pendingKey='q';
	}

	switch (key)
	{
	case 'h':
	case 'H':
		viewCollector.hierRenderMode=!viewCollector.hierRenderMode;
		printf("render mode %s\n", (viewCollector.hierRenderMode)?"hierachical":"flooding");
		break;
	case 'f':
	case 'F':
		viewCollector.fullRenderMode=!viewCollector.fullRenderMode;
		printf("full render mode is %s\n", viewCollector.fullRenderMode?"yes":"no");
		break;

	case 'm':
	case 'M':
		displayInfo();
		break;

	case '+':		
		if (octreeLoader.capacityFactor>=MaxCapacity)
		{
			octreeLoader.capacityFactor=MaxCapacity;
		}
		else
		{
			octreeLoader.capacityFactor*=2;
		}
		printf("capacityFactor=%d\n", octreeLoader.capacityFactor);
		break;
	case '-':
		octreeLoader.capacityFactor/=2;
		if (octreeLoader.capacityFactor==0)
		{
			octreeLoader.capacityFactor=1;
		}
		printf("capacityFactor=%d\n", octreeLoader.capacityFactor);
		break;
	case 27:
		//printf("escape?\n\n");
		break;//must mask this key
	default:
		renderEngine.keyboardCallback(key, x, y);
		break;

	}
	//viewCollector.startCollecting(renderEngine.drawingContext.center);
	//knapsack(viewCollector.octreeNodeVector.size(), &(*viewCollector.octreeNodeVector.begin()), false);

	glutPostRedisplay();
	
}



void MasterContext::recvNaiveVisiblePixelInfo()
{
#ifdef USING_COMPRESSION
	MPI_Status status;
	int length;
	SAFE_CALL(MPI_Irecv(compressedVisibleTable, VisibleTableSize, MPI_UNSIGNED_CHAR, TesterHostStartIndex,
		0, MPI_COMM_WORLD, &naiveVisibleTableRequest));
	SAFE_CALL(MPI_Wait(&naiveVisibleTableRequest, &status)); 
	SAFE_CALL(MPI_Get_count(&status,MPI_UNSIGNED_CHAR, &length));
	length=decompressBuffer(compressedVisibleTable, length, visibleTable, DepthBufferSize);
	assert(length==DepthBufferSize);
#else
	SAFE_CALL(MPI_Irecv(visibleTable, DepthBufferSize, MPI_UNSIGNED_CHAR, TesterHostStartIndex, 0, MPI_COMM_WORLD,
		&naiveVisibleTableRequest));
	SAFE_CALL(MPI_Wait(&naiveVisibleTableRequest, MPI_STATUS_IGNORE));
#endif
} 	




void MasterContext::recvColorBuffer()
{
	int i;	
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		//here we use single size of color buffer to hold all pixel data from all renders
		SAFE_CALL(MPI_Irecv(colorBuffer+ColorBufferAtomSize*i, ColorBufferAtomSize, MPI_UNSIGNED_CHAR, 
				MPI_ANY_SOURCE, i, MPI_COMM_WORLD, colorBufferRequests+i));
		DEBUG_INFO("displayer begin recv pixel data\n");
	}
	DEBUG_INFO("displayer post requests to receive pixel data\n");
}

void MasterContext::testColorBufferRecving()
{
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, colorBufferRequests, MPI_STATUSES_IGNORE));
}




