
/*************************************************************************************
file name: master.h
author: qingzhe huang
date: Jul. 21, 2007
This is a combination of "Master" and "Displayer" module which maybe needs to be separated
in other situation. i.e. It runs a too heavy job as a role of two.

*************************************************************************************/
#pragma once

#include "config.h"
#include "dataStructure.h"
#include "viewCollector.h"
#include "renderEngine.h"
#include <vector>

using namespace std;

struct MasterContext
{
	RenderEngine renderEngine;
	OctreeLoader octreeLoader;
	ViewCollector viewCollector;
	static char* windowTitle;
	int renderPixelCount[RenderCount+1];
	ubyte* renderPixelPtr[RenderCount];

	ubyte* colorBuffer;
	ubyte* pixelDataBuffer;
	ubyte* visibleTable;
	ubyte* testerNodeHashTable;
	ubyte* renderNodeHashTable;
	//////////////////////////////////////
	//use this as receiving buffer for compressed visible table
	ubyte* compressedVisibleTable;
	////////////////////////////////////

	MPI_Request* visibleTableRequests;
	MPI_Request* pixelDataRequests;
	MPI_Request* colorBufferRequests;
	MPI_Request naiveVisibleTableRequest;

	vector<int>* renderVector;
	vector<int>* testerVector;
	int* renderShareCount;
	int* testerShareCount;
	void init();
	void uninit();
	void doKnapsack(int count, int* array, NodeStatus nodeStatus);
	void doKnapsack(int count, int* array, int partyNumber, int* shareArray, int* shareCount, 
				vector<int>* shareVector, int* output);
	void knapsack(int count, int* array, bool doSimple=true);
	void keyboardCallback(unsigned char key, int x, int y);
	void displayInfo();
	void displayCallback();

	void recvVisiblePixelInfo();
	void testVisiblePixelInfoRecving();

	void displayerPrepareRecvPixelData();	
	void recvPixelData();
	void testPixelDataRecving();
	void recvColorBuffer();
	void testColorBufferRecving();
	void outputColorBuffer();
	void recvNaiveVisiblePixelInfo();
};

static MasterContext masterContext;


