#ifndef OCTREELOADER_H
#define OCTREELOADER_H

#include "octree.h"
#include "config.h"
#include "frustum.h"
#include <cstdio>
#include <set>

using namespace std;


typedef set<int> OctreeNodeSet;
//const int MaxVertexCapacity=512*1024;
//const int MaxFaceCapacity=512*1024;


////////////////////////////////////////////////////////////////////////////
//begin of expand by 8

struct OctreeShiftDescriptor
{
	int levelOffset;
	VVector newCenter;
	VVector relativeOps;
	int indexOffset;
	int vPageOffset;
	int fPageOffset;
	float vertexOffset;
};
void getLastError();

void vectorMultiply(VVector v1, VVector v2, VVector result);
void vectorScalor(VVector v1, VType scalor);

void vectorScalor(VVector v1, VType scalor, VVector result);

void vectorMinus(VVector v1, VVector v2, VVector result);
void vectorPlus(VVector v1, VVector v2, VVector result);

struct OctreeLoader
{
	//static HANDLE loadingThread;
	//static HANDLE mutex;
	//static HANDLE event;
	//static CRITICAL_SECTION criticalSection;
	//static bool canTerminate;
	//static int ignoredLevel;
	static HANDLE vertexMapHandle, faceMapHandle;
	//PageManager vertexManager, faceManager;
	static int currentIndex, nextIndex;
	static OctreeNodeSet nodeSet[MaxCoherentFrameNumber]; 
	//OctreeNodeQueue nodeQueue[2];
	static OctreeNodeSet loadedNodeSet;
	//static OctreeNodeQueue loadingQueue;
	//static bool usingThread;

	static unsigned int loadedVertexCounter, loadedFaceCounter;
	static int capacityFactor;

	void addCandidate(int index);
	void addCandidates(int count, int* array);

	//void printPriority();

	////////////////////////////////////////////////////////////////////////
	//this is a utility for shift-expanding of octree
	static void copyOctreeLeafNode(Octree* refOctree, Octree* targetOctree, OctreeShiftDescriptor& octreeShift);
	static void calcOctreeShiftDescriptor(OctreeShiftDescriptor& octreeShift, int i);
	static void calcNewOctreeRoot(Octree& newRoot);

	//This is a bit tricky, 
	//bool leafNodeContainPoint(int index, FVector origin, OctreeNodeQueue& result);


	//void waitLoadingThread();

	static void unmapUnusedNodes();
	static void mapOctreeNode(int index);
	static bool doMapOctreeNode(int index, bool readOnly=true);
	
	static void paintPseudoColor();

	static void expandBy8();

	static void readInfo();

	void displayOctreeLeafNode(int index);
	bool displayOctreeLeafNodeRepeat(int index, int approach);	

	//OctreeLoader();
	static void initialize(bool readOnly=true, bool fakeColor=false);
	static void uninitialize();
	static Octree* root;
	//~OctreeLoader();
	//void calculateOctree(int nodeIndex);

	void swapBuffer();
};

#endif
