
#include "config.h"
#include "dataStructure.h"
#include "render.h"
#include <mpi.h>
#include "except.h"
#include <GL/glut.h>



char* RendererContext::windowTitle="Renderer Window";

extern MPIContext mpiContext;

void RendererContext::init()
{
	int i;

	myRenderIndex=mpiContext.myRank-RenderHostStartIndex;
	//printf("render %d start initializing...\n", myRenderIndex);
	////////////////////////////////////////////////
	//buffer
	colorBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(colorBuffer);

	visibleTable=new ubyte[ColorBufferSize];
	CHECK_NEW(visibleTable);

	compressedVisibleTable=new ubyte[ColorBufferSize*VisibleTableBufferRate];
	CHECK_NEW(compressedVisibleTable);

	pixelDataBuffer=new ubyte[ColorBufferByteSize];
	CHECK_NEW(pixelDataBuffer);

	///////////////////////////////////////////////////////////////////
	//requests
	visibleTableRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(visibleTableRequests);

	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
	}

	pixelDataRequests=new MPI_Request;
	CHECK_NEW(pixelDataRequests);
	*pixelDataRequests=MPI_REQUEST_NULL;
	///////////////////////////////////////////////////////////////////
	naiveVisibleTableRequest=MPI_REQUEST_NULL;
	octreeLoader.initialize(true);
	renderEngine.init(&octreeLoader);
	renderEngine.lighting.lightEnabled=false;
#ifdef	PROFILING
	renderEngine.lighting.toggleLighting();
#endif
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadBuffer(GL_BACK);

}


#ifdef USING_DUMMY_RENDER
///////////////////////////////////////
//dummy render with no render job at all, only for testing
void RendererContext::displayCallback()
{
	int offset=0;
	DEBUG_INFO("render begins displaycallback\n");
	mpiContext.recvRenderContext();
	mpiContext.testRenderContextRecving();

	for (int i=0; i<myRenderIndex; i++)
	{
		offset+=mpiContext.renderContext->renderShare[i];
	}

	if (mpiContext.startMPI)
	{
		DEBUG_INFO("render starts..\n");
		recvVisiblePixelInfo();
		testVisiblePixelInfoRecving();

		sendPixelData();
		testPixelDataSending();
		DEBUG_INFO("render finishes\n");
	}
}
#else
void RendererContext::displayCallback()
{
	int offset=0;
	DEBUG_INFO("render begins displaycallback......\n");

	PROFILE_BEGIN(ALL);
	PROFILE_BEGIN(BMTRD);

	mpiContext.recvRenderContext();
	mpiContext.testRenderContextRecving();
	PROFILE_END(BMTRD);

	PROFILE_BEGIN(CR);
	renderEngine.keyboardCallback(mpiContext.renderContext->key, 0, 0);
	for (int i=0; i<myRenderIndex; i++)
	{
		offset+=mpiContext.renderContext->renderShare[i];
	}

	octreeLoader.addCandidates(mpiContext.renderContext->renderShare[myRenderIndex], 
		mpiContext.renderShareArray+offset);
	if (mpiContext.startMPI)
	{
		//recvVisiblePixelInfo();
	}
	octreeLoader.swapBuffer();
	if (mpiContext.startMPI)
	{
		//recvVisiblePixelInfo();
		readColorBuffer();
		PROFILE_END(CR);
		PROFILE_BEGIN(BTRD);		
		recvVisiblePixelInfo();
		testVisiblePixelInfoRecving();
		PROFILE_END(BTRD);		
		PROFILE_BEGIN(BRD);
		sendPixelData();
		testPixelDataSending();
		PROFILE_END(BRD);
	}
	PROFILE_END(ALL);
	DEBUG_INFO("render finshes displaycallback......\n");
}
#endif


void RendererContext::readColorBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, DepthBufferAtomWidth, 
				GL_RGBA, GL_UNSIGNED_BYTE, 
				colorBuffer+4*(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize);
		}
	}
}

void RendererContext::recvVisiblePixelInfo()
{
	for (int i=0; i<DepthBufferSnapShotNumber; i++)
	{
		visibleTableRequests[i]=MPI_REQUEST_NULL;
#ifndef	USING_COMPRESSION
		SAFE_CALL(MPI_Irecv(visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize, MPI_UNSIGNED_CHAR,
			MPI_ANY_SOURCE, i,  MPI_COMM_WORLD, visibleTableRequests+i));
#else
		SAFE_CALL(MPI_Irecv(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, DepthBufferAtomSize*VisibleTableBufferRate,
			 MPI_UNSIGNED_CHAR, MPI_ANY_SOURCE, i,  MPI_COMM_WORLD, visibleTableRequests+i));
#endif
	}
	DEBUG_INFO("Render receives visible pixel info of frame number... \n");
}

void RendererContext::testVisiblePixelInfoRecving()
{
#ifdef	USING_COMPRESSION
	MPI_Status status[DepthBufferSnapShotNumber];
	int i, length;
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, status));
	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		SAFE_CALL(MPI_Get_count(status+i, MPI_UNSIGNED_CHAR, &length));
		length=decompressBuffer(compressedVisibleTable+i*DepthBufferAtomSize*VisibleTableBufferRate, length, 
			visibleTable+i*DepthBufferAtomSize, DepthBufferAtomSize);
		assert(length==DepthBufferAtomSize);
	}
#else
	SAFE_CALL(MPI_Waitall(DepthBufferSnapShotNumber, visibleTableRequests, MPI_STATUSES_IGNORE));
#endif
	DEBUG_INFO("Render finishes testing pixelInfo of frame receiving\n");
}
	

void RendererContext::sendPixelData()
{
	int index=0;
	int displayerIndex;
#ifdef	USING_SEPARATE_DISPLAYER
	displayerIndex=DisplayerIndex;
#else
	displayerIndex=MasterIndex;
#endif
	for (int i=0; i<ColorBufferSize; i++)
	{
		if (visibleTable[i]==myRenderIndex)
		{
			memcpy(pixelDataBuffer+index*4, colorBuffer+i*4, 4);
			index++;
		}
	}
	*pixelDataRequests=MPI_REQUEST_NULL;
	if (index>0)
	{
		SAFE_CALL(MPI_Isend(pixelDataBuffer, index*4, MPI_UNSIGNED_CHAR, displayerIndex, PIXEL_DATA,
			MPI_COMM_WORLD,	pixelDataRequests));
	}
	DEBUG_INFO("Render sends pixel data to displayer.........\n");
}

void RendererContext::testPixelDataSending()
{
	SAFE_CALL(MPI_Wait(pixelDataRequests, MPI_STATUS_IGNORE));
	DEBUG_INFO("Render finished pixel data sending.....\n");
}


void RendererContext::recvNaiveVisiblePixelInfo()
{
#ifdef USING_COMPRESSION
        MPI_Status status;
        int length;
        SAFE_CALL(MPI_Irecv(compressedVisibleTable, VisibleTableSize, MPI_UNSIGNED_CHAR, TesterHostStartIndex,
                0, MPI_COMM_WORLD, &naiveVisibleTableRequest));
        SAFE_CALL(MPI_Wait(&naiveVisibleTableRequest, &status));
        SAFE_CALL(MPI_Get_count(&status,MPI_UNSIGNED_CHAR, &length));
        length=decompressBuffer(compressedVisibleTable, length, visibleTable, DepthBufferSize);
        assert(length==DepthBufferSize);
#else
        SAFE_CALL(MPI_Irecv(visibleTable, DepthBufferSize, MPI_UNSIGNED_CHAR, TesterHostStartIndex, 0, MPI_COMM_WORLD,
                &naiveVisibleTableRequest));
        SAFE_CALL(MPI_Wait(&naiveVisibleTableRequest, MPI_STATUS_IGNORE));
	//DEBUG_INFO("render recv naive visible pixel info\n");
#endif
}

