#include "syscall.h"

#ifdef LINUX_VERSION
#define __USE_FILE_OFFSET64 1
#include <cstdio>
#include <cstdlib>
#include <dirent.h>
#include <string>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>

using namespace std;

//typedef unsigned long long INTEGER64;

long int GenericGetTime()
{
	struct timeval tv;
	long int result;
	LINUX_SYSCALL(gettimeofday(&tv, NULL), "gettimeofday")
	result=tv.tv_sec*1000+tv.tv_usec/1000;
	return result;	
}

long int GenericGetProfileTime()
{
	struct timeval tv;
	long int result;
	LINUX_SYSCALL(gettimeofday(&tv, NULL), "gettimeofday")
	result=tv.tv_sec*1000000+tv.tv_usec;
	return result;	
}

HANDLE GenericOpenFile(char* fileName, bool readOnly)
{
	int flag=S_IWUSR|S_IRUSR|S_IROTH|S_IWOTH;
	int accessRight=readOnly?O_RDONLY|O_LARGEFILE:O_RDWR|O_CREAT;
	HANDLE result;
	if (readOnly)
	{
		result=open(fileName, accessRight);
	}
	else
	{
		result=open(fileName, accessRight, flag);
	}
	LINUX_SYSCALL(result, "open file error");
	return result;
}


int GenericWriteFile(HANDLE fileHandle, void* buffer, size_t length)
{
	int result=write(fileHandle, buffer, length);
	LINUX_SYSCALL(result, "write file error");
	return result;
}


int GenericReadFile(HANDLE fileHandle, void* buffer, size_t length)
{
	int result;
	result=read(fileHandle, buffer, length);
	LINUX_SYSCALL(result, "read file error");
	return result;
}

const char* Pattern="temp.XXXXXX";

HANDLE GenericOpenTempFile(const char* nameString, char* buf)
{
/*
	HANDLE result;
	if (!tmpnam(buf+2))
	{
		perror("create temp file name");
		exit(123);
	}
	buf[0]='.';
	buf[1]='/';
	result=open(buf, O_RDWR|O_CREAT|O_LARGEFILE);
	LINUX_SYSCALL(result, "open temp file");
	return result;
*/
	HANDLE result;
	strcpy(buf, Pattern);
	result=mkstemp(buf);
	LINUX_SYSCALL(result, "create temp file");
	return result;
}

void GenericCloseFile(HANDLE fileHandle)
{
	LINUX_SYSCALL(close(fileHandle), "close file");
}

void GenericCloseTempFile(HANDLE fileHandle, const char* fileName)
{
	LINUX_SYSCALL(close(fileHandle), "close file");
	LINUX_SYSCALL(unlink(fileName), "unlink temp file");
}

void GenericSetFilePointer(HANDLE fileHandle, unsigned long int low, unsigned long int high)
{
	unsigned long long offset=0;
	offset=high;
	offset<<=32;
	offset+=low;
	LINUX_SYSCALL(lseek(fileHandle, offset, SEEK_SET), "set file pointer");
}

HANDLE GenericCreateFileMapping(HANDLE fileHandle, bool readOnly)
{
	return fileHandle;
}


void* GenericMapViewOfFile(HANDLE fileMappingHandle, bool readOnly, unsigned long int fileAddHigh, unsigned long int 
fileAddLow, unsigned int length)
{
	void* result;
	int accessRight=readOnly?PROT_READ:PROT_READ|PROT_WRITE; 
	LONGLONG offset=0;
	offset=fileAddHigh;
	offset<<=32;
	offset+=fileAddLow;
	//offset>>=12;//this is because mmap2 uses 4k as offset to represent large file.
	//this mmap2 must be Unix instead of linux
	result=mmap(0, length, accessRight, MAP_SHARED, fileMappingHandle, offset);
	//LINUX_SYSCALL(result, "map view of file");
	if (result==NULL)
	{
		perror("map view of file");
		exit(1234);
	}
	return result;
}


void GenericUnmapViewOfFile(void* address, int length)
{
	LINUX_SYSCALL(munmap(address, length), "unmap view of file");
}



void GenericRetrieveAllFile(char* dirName, vector<string>& fileNameVector)
{
	struct stat buf;	
	char curFileName[256];
	int nameLen;

	DIR* dirPtr;
	struct dirent* direntPtr;
	
	LINUX_SYSCALL(lstat(dirName, &buf), "read stat");
	
	
	if (S_ISREG(buf.st_mode))
	{
		fileNameVector.push_back(string(dirName));
	}
	else
	{
		if (S_ISDIR(buf.st_mode))
		{			
			if ((dirPtr=opendir(dirName))==NULL)
			{
				perror("read dir");
				exit(1234);
			}
			nameLen=strlen(dirName);
			strcpy(curFileName, dirName);
			while ((direntPtr=readdir(dirPtr))!=NULL)
			{
				if (strcmp(direntPtr->d_name, ".")!=0&&strcmp(direntPtr->d_name, "..")!=0)
				{
					curFileName[nameLen]='/';
					strcpy(curFileName+nameLen+1, direntPtr->d_name);
					GenericRetrieveAllFile(curFileName, fileNameVector);
				}
			}
		}
	}
}

void GenericCloseFileMapping(HANDLE fileMappingHandle)
{
	//do nothing, actually this is only to suit windows api.
}

#endif
