#include "syscall.h"
#include <windows.h>

#ifndef LINUX_VERSION

//typedef __int64 INTEGER64;

const int MaxTempNameStringLength=8+4+1;

unsigned long int temp;

char tempName[MaxTempNameStringLength]=

{

	'A','A','A','A','A','A','A','A','.','t','m','p', '\0'
};


long int GenericGetTime()
{
	return GetTickCount();
}

HANDLE GenericOpenFile(char* fileName, bool readOnly)
{
	DWORD accessRight=readOnly?GENERIC_READ:GENERIC_READ|GENERIC_WRITE;
	DWORD createDisposition=readOnly?OPEN_EXISTING:OPEN_ALWAYS;
	HANDLE result;

	result=CreateFile(fileName, accessRight, FILE_SHARE_READ, NULL, createDisposition, 
		FILE_ATTRIBUTE_NORMAL, NULL);
	WINAPI_CALL(result, "open file");

	return result;
}


int GenericWriteFile(HANDLE fileHandle, void* buffer, size_t length)
{
	WINAPI_CALL(WriteFile(fileHandle, buffer, length, &temp, NULL), "write file error");
	return temp;
}

//Linux "read" function is more sensible in indicating end of file by returning 0
//and -1 for error. windows is non-sense in this context.
int GenericReadFile(HANDLE fileHandle, void* buffer, size_t length)
{
	int result=ReadFile(fileHandle, buffer, length, &temp, NULL);
	WINAPI_CALL(result, "read file");

	return temp;
}

//this is dirty because windows API "GetTempFileName" is written by a naive guy who 
//believe people only generate temp file of total number of 65535
//while in Linux, they are better, because they all 26^6 which is roughly hundreds of million
HANDLE GenericOpenTempFile(const char* nameString, char* buf)
{
	HANDLE result;
	int pos=0;
	while (true)
	{
		if (tempName[pos]=='Z')
		{
			tempName[pos]='A';
			pos++;
		}
		else
		{
			tempName[pos]++;
			break;
		}
	}
	strcpy(buf, nameString);
	strcat(buf, tempName);

	//WINAPI_CALL(GetTempFileName(".", nameString, 0, buf), "create temp file name");

	//result=CreateFile(buf, GENERIC_READ|GENERIC_WRITE, 0, NULL, 
	//			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	result=CreateFile(buf, GENERIC_READ|GENERIC_WRITE, 0, NULL, 
				CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
	WINAPI_CALL(result, "create temp file");
	return result;
}


void GenericCloseFile(HANDLE fileHandle)
{
	WINAPI_CALL(CloseHandle(fileHandle), "close file");

}



void GenericCloseTempFile(HANDLE fileHandle, const char* fileName)
{
	WINAPI_CALL(CloseHandle(fileHandle), "close temp file");
	WINAPI_CALL(DeleteFile(fileName), "delete temp file");
}


void GenericSetFilePointer(HANDLE fileHandle, long int low, long int high)
{
	if (SetFilePointer(fileHandle, low, &high, FILE_BEGIN)==0xFFFFFFFF)
	{
		printf("error of %s #%d\n", "set file pointer", GetLastError()); 
		exit(123);
	}
}


HANDLE GenericCreateFileMapping(HANDLE fileHandle, bool readOnly)
{
	DWORD accessRight=readOnly?PAGE_READONLY:PAGE_READWRITE;
	HANDLE result;
	result=CreateFileMapping(fileHandle, NULL, accessRight, 0, 0, NULL);

	WINAPI_CALL(result, "create file mapping");
	return result;
}

void GenericCloseFileMapping(HANDLE fileMappingHandle)
{
	WINAPI_CALL(CloseHandle(fileMappingHandle), "close file mapping");
}

void* GenericMapViewOfFile(HANDLE fileMappingHandle, bool readOnly, int fileAddHigh, int fileAddLow, int length)
{
	DWORD accessRight=readOnly?FILE_MAP_READ:FILE_MAP_WRITE;
	void*result;
	result=MapViewOfFile(fileMappingHandle, accessRight, fileAddHigh, fileAddLow, length);
	WINAPI_CALL(result, "map file");
	return result;
}


void GenericUnmapViewOfFile(void* address, int length)
{
	WINAPI_CALL(UnmapViewOfFile(address), "unmap view of file");
}



void GenericRetrieveAllFile(char* dir, vector<string>& fileNameVector)
{
	HANDLE handle;	
	char curFileName[256];
	char wildFileName[256];
	
	WIN32_FIND_DATA ffd;
	
	sprintf(wildFileName, "%s\\*.*", dir);
	handle=FindFirstFile(wildFileName, &ffd);
	
	if (handle==INVALID_HANDLE_VALUE)
	{
		printf("findfirst failed of error code =%d\n", GetLastError());
		exit(19);
	}

	while (FindNextFile(handle, &ffd))
	{		
		if (strcmp(ffd.cFileName, "..")!=0)
		{
			sprintf(curFileName, "%s\\%s", dir, ffd.cFileName);
			if  (GetFileAttributes(curFileName)&FILE_ATTRIBUTE_DIRECTORY)
			{
				GenericRetrieveAllFile(curFileName, fileNameVector);
			}
			else
			{
				fileNameVector.push_back(string(curFileName));
			}
		}			
	}
	FindClose(handle);
}

#endif
