#include "syscall.h"
#include "config.h"
#include "dataStructure.h"
#include "tester.h"
#include "octree.h"
#include "occluder.h"
#include <cstdlib>
#include <cstdio>
#include <mpi.h>
#include <GL/glut.h>
#include <assert.h>


#ifdef DEBUGGING
#ifdef LINUX_VERSION
#include <syslog.h>

#endif
#endif

char* TesterContext::windowTitle="Tester Node";
extern MPIContext mpiContext;
extern FILE* dumpPtr;

//assume TileCount and TesterCount both being power of 2!!!!!!!!!
void BinarySwapDescriptor::init(int testerIndex)
{
	myTesterIndex=testerIndex;
	mask=0xFFFFFFFE;//32BIT INT, 64BIT OS NEED UPDATE
	//equivalent to testerIndex/2*2
	myLeaderIndex=testerIndex&mask;
	myGroupSize=2;
	myTileCount=DepthBufferSnapShotNumber/2;
	if (myTesterIndex==myLeaderIndex)
	{
		myTileIndex=0;
		mySendIndex=myTileIndex+myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=true;
		myBuddyIndex=myLeaderIndex+1;
	}
	else
	{
		myTileIndex=DepthBufferSnapShotNumber/2;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=false;
		myBuddyIndex=myLeaderIndex;
	}
}
	
bool BinarySwapDescriptor::next()
{
	mask<<=1;
	if ((mask&TesterCount)==0)
	{
		return false;
	}
	myLeaderIndex=mask&myTesterIndex;
	myTileCount/=2;
	if (myTesterIndex<myLeaderIndex+myGroupSize)
	{
		myBuddyIndex=myTesterIndex+myGroupSize;
		myRecvIndex=myTileIndex;
		mySendIndex=myTileIndex+myTileCount;
	}
	else
	{
		myBuddyIndex=myTesterIndex-myGroupSize;
		myTileIndex+=myTileCount;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
	}
	myGroupSize*=2;
	return true;
}		


void BinarySwapDescriptor::display()
{
	
	printf("\n******************************\nBinarySwapDescriptor: myTesterIndex=%d, myLeaderIndex=%d,myGroupSize=%d,\
		\nmyBuddyIndex=%d,myTileIndex=%d,myTileCount=%d,\n\
		mySendIndex=%d,myRecvIndex=%d,mask=%X\n", myTesterIndex, myLeaderIndex, myGroupSize,myBuddyIndex,\ 
		myTileIndex, myTileCount, mySendIndex, myRecvIndex, mask);
}


void TesterContext::init()
{
	int i, j;
	MPI_Group commGroup;
	int testerRanks[TesterCount];

	myTesterIndex=mpiContext.myRank-TesterHostStartIndex;
	depthBufferCount=2;
	colorBufferCount=2;

	colorBuffer=new ubyte*[colorBufferCount];
	CHECK_NEW(colorBuffer);

	for (i=0; i<colorBufferCount; i++)
	{
		colorBuffer[i]=new ubyte[ColorBufferByteSize];
		CHECK_NEW(colorBuffer[i]);
	}
	myColorBuffer=colorBuffer[0];
	hisColorBuffer=colorBuffer[1];
	depthBuffer=new float*[depthBufferCount];
	CHECK_NEW(depthBuffer);

	for (i=0; i<depthBufferCount; i++)
	{
		depthBuffer[i]=new float[DepthBufferSize];
		CHECK_NEW(depthBuffer[i]);
	}	
	myDepthBuffer=depthBuffer[0];
	hisDepthBuffer=depthBuffer[1];


	binarySwapDescriptor=new BinarySwapDescriptor;
	CHECK_NEW(binarySwapDescriptor);



	depthBufferExchangeRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(depthBufferExchangeRequests);
	colorBufferExchangeRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(colorBufferExchangeRequests);

	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		depthBufferExchangeRequests[i]=colorBufferExchangeRequests[i]=MPI_REQUEST_NULL;
	}

	for (i=0; i<TesterCount; i++)
	{
		testerRanks[i]=TesterHostStartIndex+i;
	}

	SAFE_CALL(MPI_Comm_group(MPI_COMM_WORLD, &commGroup));
	SAFE_CALL(MPI_Group_incl(commGroup, TesterCount, testerRanks, &testerGroup));
	SAFE_CALL(MPI_Comm_create(MPI_COMM_WORLD, testerGroup, &testerComm));

	octreeLoader.initialize(true);
	renderEngine.init(&octreeLoader);
	renderEngine.lighting.toggleLighting();

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
//////////////////////////////////////////////////////
	//occluder init
	occluder.init(myTesterIndex);
////////////////////////////////////////////////
	glReadBuffer(GL_BACK);

#ifdef 	DEBUGGING
#ifdef LINUX_VERSION
	openlog("renderBox", 0,0);
#endif
#endif


}



void TesterContext::keyboardCallback(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 0:
		break;
	default:
		renderEngine.keyboardCallback(key, x, y);
		break;
	}
	glutPostRedisplay();
}


void TesterContext::displayCallback()
{
	int offset;
	DEBUG_INFO("tester begins displaycallback......\n");

	PROFILE_BEGIN(ALL);
	
	PROFILE_BEGIN(BMTRD);
	mpiContext.recvRenderContext();

	mpiContext.testRenderContextRecving();
	PROFILE_END(BMTRD);


	PROFILE_BEGIN(CT);
	////////////////////////////////////////////////
	keyboardCallback(mpiContext.renderContext->key, 0, 0);
	offset=0;
	for (int i=0; i<myTesterIndex; i++)
	{
		offset+=mpiContext.renderContext->testerShare[i];
	}
	octreeLoader.addCandidates(mpiContext.renderContext->testerShare[myTesterIndex], 
		mpiContext.testerShareArray+offset);

	if (mpiContext.renderContext->frameNumber==80)
	{
		occluderTest();
	}
	else
	{
		octreeLoader.swapBuffer();
	}

	if (mpiContext.startMPI)
	{		
		DEBUG_INFO("tester starts...\n");
		readColorBuffer();
		readDepthBuffer();
		PROFILE_END(CT);
		PROFILE_BEGIN(BT);
		exchangeDepthBuffer();
		PROFILE_END(BT);
		PROFILE_BEGIN(BRD);
		//testDepthBufferExchanging();
		sendColorBuffer();
		testColorBufferSending();
		PROFILE_END(BRD);
	}
	PROFILE_END(ALL);
	DEBUG_INFO("tester finishes displaycallback......\n");
}

void TesterContext::debugPrint(const char* msg)
{
	fprintf(occluder.log, "%s", msg);
	fflush(occluder.log);
}

#if 0
void TesterContext::occluderTest()
{
	int i=0, counter=0, exchangeCounter=0;
	Rect rect;
	Rect fullScreen;
	float depth;
	GLenum err;
	bool finished;
	bool canStop=false;
	int approach;
        OctreeNodeSet::iterator it;
	occluder.calcTransform();
        octreeLoader.unmapUnusedNodes();
	int faceNumber;
	int startIndex;
	int totalTry;
	int succeededTry;
	int failedTry;
	int number;
	PlyVertex* vPtr;
	UIVector* fPtr;
	fullScreen.topLeft.x=fullScreen.topLeft.y=0;
	fullScreen.bottomRight.x=fullScreen.bottomRight.y=ScreenSize-1;
 	glEnableClientState (GL_VERTEX_ARRAY);
        glEnableClientState (GL_NORMAL_ARRAY);
        glEnableClientState (GL_COLOR_ARRAY);
	fprintf(occluder.log, "total octree leaf nodes %d\n", 
		octreeLoader.nodeSet[octreeLoader.nextIndex].size());
	fflush(occluder.log);

        //printf("nextset size=%d\n", nodeSet[nextIndex].size());
	totalTry=succeededTry=failedTry=0;
        for (it=octreeLoader.nodeSet[octreeLoader.nextIndex].begin(); 
		it!=octreeLoader.nodeSet[octreeLoader.nextIndex].end(); it++)
        {
		approach = 0;
		finished = false;
		printf("rank[%d]node[%d]\n", myTesterIndex, *it);
		while (!canStop && !finished)
		{
			//if (!occluder.calcRegion(octreeLoader.root[*it].center, 
			//	octreeLoader.root[*it].width/2, rect, depth))
			startIndex = approach*DisplayOctreeLeafNodeBatchNumber;
			faceNumber = octreeLoader.root[*it].fNumber - startIndex;
			approach ++;
			if (octreeLoader.root[*it].fNumber <= 0)
			{
				printf("fNumber <= 0\n\n");
				break;
			}
			faceNumber = MIN(MAX(0, faceNumber), DisplayOctreeLeafNodeBatchNumber);
			//printf("faceNumber = %d\n", octreeLoader.root[*it].fNumber);
			if (faceNumber == 0)
			{
				printf("faceNumber == 0\n");	
				break;
			}
			vPtr = (PlyVertex*)(octreeLoader.root[*it].vAddress);
			fPtr = (UIVector*)(octreeLoader.root[*it].fAddress);
			if (!occluder.calcRegion(vPtr, fPtr+startIndex, faceNumber, rect, depth))
			{
				fprintf(occluder.log, "calc region failed node[%d], faceNumber[%d]\n", myTesterIndex, faceNumber);
				fflush(occluder.log);
				finished = true;
				continue;
			}
			if (!occluder.exchangeRegion())
			{
				canStop=true;
			}
			else
			{
				occluder.mergeDepthBuffer();
		
				if (occluder.checkRegion(rect, depth))
				{
					fprintf(occluder.log, "#[%d]rect[%d,%d][%d,%d]depth[%f]excluded\n", 
							counter, rect.topLeft.x, rect.topLeft.y, rect.bottomRight.x, 
							rect.bottomRight.y, depth);
					counter++;			
					succeededTry++;
				}
				else
				{
					failedTry++;
					fprintf(occluder.log, "#[%d]rect[%d,%d][%d,%d]depth[%f] failed to excluded\n", 
							counter, rect.topLeft.x, rect.topLeft.y, rect.bottomRight.x, 
							rect.bottomRight.y, depth);
					//occluder.printDepthBuffer(rect);
				}
			}
		
        	       	//octreeLoader.displayOctreeLeafNode(*it);
			totalTry++;
			if (octreeLoader.displayOctreeLeafNodeRepeat(*it, approach))
			{
				finished = true;
			}
		}

		i++;
		if (canStop)
		{
			break;
		}	
        }
	if (!canStop)
	{
		occluder.sendEndExchange();
	}
	fprintf(occluder.log, "totalTry=[%d], succeeded=[%d], failed=[%d]\n", totalTry, succeededTry, failedTry);

	fprintf(occluder.log, "total octree leaf nodes %d, occluded %d\n", 
		octreeLoader.nodeSet[octreeLoader.nextIndex].size(), counter);
	fprintf(occluder.log, "regionset size=%d, exchangeCounter=%d\n", occluder.regionSet.size(), 
		exchangeCounter);
	//occluder.outputRegionSet();
	fclose(occluder.log); 
        glDisableClientState (GL_VERTEX_ARRAY);
        glDisableClientState (GL_NORMAL_ARRAY);
        glDisableClientState (GL_COLOR_ARRAY);

        octreeLoader.currentIndex=octreeLoader.nextIndex;
        octreeLoader.nextIndex=(octreeLoader.nextIndex+1)%MaxCoherentFrameNumber;
        octreeLoader.nodeSet[octreeLoader.nextIndex].clear();
}
#endif

//int TesterContext::doOccluderTest(PlyVertex* vPtr, UIVector* fPtr, int faceNumber)
//{
	

void TesterContext::occluderTest()
{
	bool canStop=false;
        OctreeNodeSet::iterator it;
	int faceNumber;
	int startIndex;
	int totalTry;
	int succeededTry;
	int number;
	int counter = 0;
	PlyVertex* vPtr;
	UIVector* fPtr;


	occluder.calcTransform();
        octreeLoader.unmapUnusedNodes();

 	glEnableClientState (GL_VERTEX_ARRAY);
        glEnableClientState (GL_NORMAL_ARRAY);
        glEnableClientState (GL_COLOR_ARRAY);

	fprintf(occluder.log, "total octree leaf nodes %d\n", octreeLoader.nodeSet[octreeLoader.nextIndex].size());
	fflush(occluder.log);

	totalTry = succeededTry = 0;
        for (it=octreeLoader.nodeSet[octreeLoader.nextIndex].begin(); 
		it!=octreeLoader.nodeSet[octreeLoader.nextIndex].end(); it++)
        {
		//fprintf(occluder.log, "counter[%d]\n", counter++);
		
		faceNumber = octreeLoader.root[*it].fNumber;
		vPtr = (PlyVertex*)(octreeLoader.root[*it].vAddress);
		fPtr = (UIVector*)(octreeLoader.root[*it].fAddress);
		number = occluder.doOccluderTest(vPtr, fPtr, faceNumber);
		fprintf(occluder.log, "node[%d]faceNumber[%d]occluded[%d]\n", *it, faceNumber, number);
		totalTry  += faceNumber;
		succeededTry += number;
		fflush(occluder.log);

       	       	octreeLoader.displayOctreeLeafNode(*it);
		if (!occluder.exchangeRegion())
		{
			canStop = true;
			break;
		}
		occluder.mergeDepthBuffer();
        }
	occluder.checkDepthBuffer();
	if (!canStop)
	{
		occluder.sendEndExchange();
	}
	fprintf(occluder.log, "totalTry=[%d], succeeded=[%d]\n", totalTry, succeededTry);

	fclose(occluder.log); 
        glDisableClientState (GL_VERTEX_ARRAY);
        glDisableClientState (GL_NORMAL_ARRAY);
        glDisableClientState (GL_COLOR_ARRAY);

        octreeLoader.currentIndex=octreeLoader.nextIndex;
        octreeLoader.nextIndex=(octreeLoader.nextIndex+1)%MaxCoherentFrameNumber;
        octreeLoader.nodeSet[octreeLoader.nextIndex].clear();
}


void TesterContext::readColorBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, 
			DepthBufferAtomWidth, GL_RGBA, GL_UNSIGNED_BYTE, 
			myColorBuffer+(i*DepthBufferSnapShotWidth+j)*ColorBufferAtomSize);
		}
	}
}

void TesterContext::readDepthBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, 
				DepthBufferAtomWidth, GL_DEPTH_COMPONENT, GL_FLOAT, 
				myDepthBuffer+(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize);
		}
	}
}


void TesterContext::exchangeDepthBuffer()
{
	int length, i;
	int depthSize;
	int colorSize;
	int buddyIndex;
	float* depthSendPtr, *depthRecvPtr;
	ubyte* colorSendPtr, *colorRecvPtr;
	float* myOwnDPtr;
	ubyte* myOwnCPtr;
	//ubyte* myCPtr, *hisCPtr;
	//float* myDPtr, *hisDPtr;
	MPI_Status status;

	if (TesterCount==1)
	{
		return;
	}
	binarySwapDescriptor->init(myTesterIndex);
	do
	{
		depthSize=binarySwapDescriptor->myTileCount*DepthBufferAtomSize;
		colorSize=binarySwapDescriptor->myTileCount*ColorBufferAtomSize;

		depthRecvPtr=hisDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		depthSendPtr=myDepthBuffer+binarySwapDescriptor->mySendIndex*DepthBufferAtomSize;
		
		colorRecvPtr=hisColorBuffer+binarySwapDescriptor->myRecvIndex*ColorBufferAtomSize;
		colorSendPtr=myColorBuffer+binarySwapDescriptor->mySendIndex*ColorBufferAtomSize;

		buddyIndex=TesterHostStartIndex+binarySwapDescriptor->myBuddyIndex;

		

		SAFE_CALL(MPI_Irecv(depthRecvPtr, depthSize, MPI_FLOAT, buddyIndex,
			BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex));

		SAFE_CALL(MPI_Irecv(colorRecvPtr, colorSize, MPI_UNSIGNED_CHAR, buddyIndex,
			BINARY_SWAP_COLOR_BUFFER, MPI_COMM_WORLD, colorBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex));


		SAFE_CALL(MPI_Isend(depthSendPtr, depthSize, MPI_FLOAT, buddyIndex, BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, 
			depthBufferExchangeRequests+myTesterIndex));

		SAFE_CALL(MPI_Isend(colorSendPtr, colorSize, MPI_UNSIGNED_CHAR, buddyIndex, BINARY_SWAP_COLOR_BUFFER, 
			MPI_COMM_WORLD, colorBufferExchangeRequests+myTesterIndex));

		 		
		DEBUG_INFO("recving binary swap\n");
		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex, MPI_STATUS_IGNORE));
		SAFE_CALL(MPI_Wait(colorBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex, MPI_STATUS_IGNORE));
		

		//mySend is what I send, myRecv is what I recv, but the same index is used for buffer remains to compared with recved
		myOwnDPtr=myDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		myOwnCPtr=myColorBuffer+binarySwapDescriptor->myRecvIndex*ColorBufferAtomSize;

		for (i=0; i<depthSize; i++)
		{
			if (depthRecvPtr[i]<myOwnDPtr[i])
			{
				memcpy(myOwnCPtr+i*4, colorRecvPtr+i*4, 4);
				//myOwnCPtr[i]=colorRecvPtr[i];
				myOwnDPtr[i]=depthRecvPtr[i];
			}
		}

		DEBUG_INFO("sending binary swap\n");

		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));
		SAFE_CALL(MPI_Wait(colorBufferExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));
		//debugging...	
		//memset(hisDepthBuffer, 0, DepthBufferSize*sizeof(float));
		//memset(hisVisibleTable, 0, VisibleTableSize);
	}
	while (binarySwapDescriptor->next());
}



//assuming TesterCount, TileCount are all power of 2
void TesterContext::sendColorBuffer()
{
	int i, j, tileIndex;
	ubyte* colorBufferSendPtr;

	int displayerIndex;
#ifdef	USING_SEPARATE_DISPLAYER
	displayerIndex=DisplayerIndex;
#else
	displayerIndex=MasterIndex;
#endif

	tileIndex=binarySwapDescriptor->myRecvIndex;

	assert(tileIndex>=0&&tileIndex<DepthBufferSnapShotNumber);


	colorBufferSendPtr=myColorBuffer+tileIndex*ColorBufferAtomSize;

	//assert(binarySwapDescriptor->myTileCount==1);
	for (i=0; i<binarySwapDescriptor->myTileCount; i++)
	{		
		SAFE_CALL(MPI_Isend(colorBufferSendPtr+i*ColorBufferAtomSize, ColorBufferAtomSize,
			MPI_UNSIGNED_CHAR, displayerIndex, tileIndex+i, MPI_COMM_WORLD, colorBufferExchangeRequests+tileIndex+i));
	}
	//printf("Tester%d sends visible pixelInfo of total %d\n", myTesterIndex, counter);
}



void TesterContext::testColorBufferSending()
{
	int i;
	//printf("*********wait for #[%u] starting from [%u]*******\n", binarySwapDescriptor->myTileCount, binarySwapDescriptor->myRecvIndex);
	SAFE_CALL(MPI_Waitall(binarySwapDescriptor->myTileCount, colorBufferExchangeRequests+binarySwapDescriptor->myRecvIndex, 
		MPI_STATUSES_IGNORE));

	//printf("###Tester%d finishes sending visible pixelInfo\n", myTesterIndex);
}







			
			












