#include "config.h"
#include "dataStructure.h"
#include "tester.h"
#include <cstdlib>
#include <cstdio>
#include <mpi.h>
#include <GL/glut.h>
#include <assert.h>


#ifdef DEBUGGING
#ifdef LINUX_VERSION
#include <syslog.h>

#endif
#endif

char* TesterContext::windowTitle="Tester Node";
extern MPIContext mpiContext;
extern FILE* dumpPtr;





//assume TileCount and TesterCount both being power of 2!!!!!!!!!
void BinarySwapDescriptor::init(int testerIndex)
{
	myTesterIndex=testerIndex;
	mask=0xFFFFFFFE;//32BIT INT, 64BIT OS NEED UPDATE
	//equivalent to testerIndex/2*2
	myLeaderIndex=testerIndex&mask;
	myGroupSize=2;
	myTileCount=DepthBufferSnapShotNumber/2;
	if (myTesterIndex==myLeaderIndex)
	{
		myTileIndex=0;
		mySendIndex=myTileIndex+myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=true;
		myBuddyIndex=myLeaderIndex+1;
	}
	else
	{
		myTileIndex=DepthBufferSnapShotNumber/2;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
		//isLow=false;
		myBuddyIndex=myLeaderIndex;
	}
}
	
bool BinarySwapDescriptor::next()
{
	mask<<=1;
	if ((mask&TesterCount)==0)
	{
		return false;
	}
	myLeaderIndex=mask&myTesterIndex;
	myTileCount/=2;
	if (myTesterIndex<myLeaderIndex+myGroupSize)
	{
		myBuddyIndex=myTesterIndex+myGroupSize;
		myRecvIndex=myTileIndex;
		mySendIndex=myTileIndex+myTileCount;
	}
	else
	{
		myBuddyIndex=myTesterIndex-myGroupSize;
		myTileIndex+=myTileCount;
		mySendIndex=myTileIndex-myTileCount;
		myRecvIndex=myTileIndex;
	}
	myGroupSize*=2;
	return true;
}		


void BinarySwapDescriptor::display()
{
	
	printf("\n******************************\nBinarySwapDescriptor: myTesterIndex=%d, myLeaderIndex=%d,myGroupSize=%d,\
		\nmyBuddyIndex=%d,myTileIndex=%d,myTileCount=%d,\n\
		mySendIndex=%d,myRecvIndex=%d,mask=%X\n", myTesterIndex, myLeaderIndex, myGroupSize,myBuddyIndex,\ 
		myTileIndex, myTileCount, mySendIndex, myRecvIndex, mask);
}


void TesterContext::init()
{
	int i, j;
	MPI_Group commGroup;
	int testerRanks[TesterCount];

	myTesterIndex=mpiContext.myRank-TesterHostStartIndex;
	depthBufferCount=2;
	colorBufferCount=2;

	colorBuffer=new ubyte*[colorBufferCount];
	CHECK_NEW(colorBuffer);

	for (i=0; i<colorBufferCount; i++)
	{
		colorBuffer[i]=new ubyte[ColorBufferByteSize];
		CHECK_NEW(colorBuffer[i]);
	}
	myColorBuffer=colorBuffer[0];
	hisColorBuffer=colorBuffer[1];
	depthBuffer=new float*[depthBufferCount];
	CHECK_NEW(depthBuffer);

	for (i=0; i<depthBufferCount; i++)
	{
		depthBuffer[i]=new float[DepthBufferSize];
		CHECK_NEW(depthBuffer[i]);
	}	
	myDepthBuffer=depthBuffer[0];
	hisDepthBuffer=depthBuffer[1];


	binarySwapDescriptor=new BinarySwapDescriptor;
	CHECK_NEW(binarySwapDescriptor);



	depthBufferExchangeRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(depthBufferExchangeRequests);
	colorBufferExchangeRequests=new MPI_Request[DepthBufferSnapShotNumber];
	CHECK_NEW(colorBufferExchangeRequests);

	for (i=0; i<DepthBufferSnapShotNumber; i++)
	{
		depthBufferExchangeRequests[i]=colorBufferExchangeRequests[i]=MPI_REQUEST_NULL;
	}

	for (i=0; i<TesterCount; i++)
	{
		testerRanks[i]=TesterHostStartIndex+i;
	}

	SAFE_CALL(MPI_Comm_group(MPI_COMM_WORLD, &commGroup));
	SAFE_CALL(MPI_Group_incl(commGroup, TesterCount, testerRanks, &testerGroup));
	SAFE_CALL(MPI_Comm_create(MPI_COMM_WORLD, testerGroup, &testerComm));

#ifdef	USING_TESTER_RENDER
	octreeLoader.initialize(true);
#else
	octreeLoader.initialize(true, true);
#endif
	renderEngine.init(&octreeLoader);
	renderEngine.lighting.toggleLighting();

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadBuffer(GL_BACK);

#ifdef 	DEBUGGING
#ifdef LINUX_VERSION
	openlog("renderBox", 0,0);
#endif
#endif


}



void TesterContext::keyboardCallback(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 0:
		break;
	default:
		renderEngine.keyboardCallback(key, x, y);
		break;
	}
	glutPostRedisplay();
}


void TesterContext::displayCallback()
{
	int offset;
	DEBUG_INFO("tester begins displaycallback......\n");

	PROFILE_BEGIN(ALL);
	
	PROFILE_BEGIN(BMTRD);
	mpiContext.recvRenderContext();
	mpiContext.testRenderContextRecving();
	PROFILE_END(BMTRD);


	PROFILE_BEGIN(CT);
	////////////////////////////////////////////////
	keyboardCallback(mpiContext.renderContext->key, 0, 0);
	offset=0;
	for (int i=0; i<myTesterIndex; i++)
	{
		offset+=mpiContext.renderContext->testerShare[i];
	}
	octreeLoader.addCandidates(mpiContext.renderContext->testerShare[myTesterIndex], 
		mpiContext.testerShareArray+offset);
	octreeLoader.swapBuffer();
	if (mpiContext.startMPI)
	{		
		DEBUG_INFO("tester starts...\n");
		readColorBuffer();
		readDepthBuffer();
		PROFILE_END(CT);
		PROFILE_BEGIN(BT);
		exchangeDepthBuffer();
		PROFILE_END(BT);
		PROFILE_BEGIN(BRD);
		//testDepthBufferExchanging();
		sendColorBuffer();
		testColorBufferSending();
		PROFILE_END(BRD);
	}
	PROFILE_END(ALL);
	DEBUG_INFO("tester finishes displaycallback......\n");
}

void TesterContext::readColorBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, 
			DepthBufferAtomWidth, GL_RGBA, GL_UNSIGNED_BYTE, 
			myColorBuffer+(i*DepthBufferSnapShotWidth+j)*ColorBufferAtomSize);
		}
	}
}

void TesterContext::readDepthBuffer()
{
	for (int i=0; i<DepthBufferSnapShotWidth; i++)
	{
		for (int j=0; j<DepthBufferSnapShotWidth; j++)
		{
			glReadPixels(i*DepthBufferAtomWidth, j*DepthBufferAtomWidth, DepthBufferAtomWidth, 
				DepthBufferAtomWidth, GL_DEPTH_COMPONENT, GL_FLOAT, 
				myDepthBuffer+(i*DepthBufferSnapShotWidth+j)*DepthBufferAtomSize);
		}
	}
}


void TesterContext::exchangeDepthBuffer()
{
	int length, i;
	int depthSize;
	int colorSize;
	int buddyIndex;
	float* depthSendPtr, *depthRecvPtr;
	ubyte* colorSendPtr, *colorRecvPtr;
	float* myOwnDPtr;
	ubyte* myOwnCPtr;
	//ubyte* myCPtr, *hisCPtr;
	//float* myDPtr, *hisDPtr;
	MPI_Status status;

	if (TesterCount==1)
	{
		return;
	}
	binarySwapDescriptor->init(myTesterIndex);
	do
	{
		depthSize=binarySwapDescriptor->myTileCount*DepthBufferAtomSize;
		colorSize=binarySwapDescriptor->myTileCount*ColorBufferAtomSize;

		depthRecvPtr=hisDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		depthSendPtr=myDepthBuffer+binarySwapDescriptor->mySendIndex*DepthBufferAtomSize;
		
		colorRecvPtr=hisColorBuffer+binarySwapDescriptor->myRecvIndex*ColorBufferAtomSize;
		colorSendPtr=myColorBuffer+binarySwapDescriptor->mySendIndex*ColorBufferAtomSize;

		buddyIndex=TesterHostStartIndex+binarySwapDescriptor->myBuddyIndex;

		

		SAFE_CALL(MPI_Irecv(depthRecvPtr, depthSize, MPI_FLOAT, buddyIndex,
			BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex));

		SAFE_CALL(MPI_Irecv(colorRecvPtr, colorSize, MPI_UNSIGNED_CHAR, buddyIndex,
			BINARY_SWAP_COLOR_BUFFER, MPI_COMM_WORLD, colorBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex));


		SAFE_CALL(MPI_Isend(depthSendPtr, depthSize, MPI_FLOAT, buddyIndex, BINARY_SWAP_DEPTH_BUFFER, MPI_COMM_WORLD, 
			depthBufferExchangeRequests+myTesterIndex));

		SAFE_CALL(MPI_Isend(colorSendPtr, colorSize, MPI_UNSIGNED_CHAR, buddyIndex, BINARY_SWAP_COLOR_BUFFER, 
			MPI_COMM_WORLD, colorBufferExchangeRequests+myTesterIndex));

		 		
		DEBUG_INFO("recving binary swap\n");
		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex, MPI_STATUS_IGNORE));
		SAFE_CALL(MPI_Wait(colorBufferExchangeRequests+binarySwapDescriptor->myBuddyIndex, MPI_STATUS_IGNORE));
		

		//mySend is what I send, myRecv is what I recv, but the same index is used for buffer remains to compared with recved
		myOwnDPtr=myDepthBuffer+binarySwapDescriptor->myRecvIndex*DepthBufferAtomSize;
		myOwnCPtr=myColorBuffer+binarySwapDescriptor->myRecvIndex*ColorBufferAtomSize;

		for (i=0; i<depthSize; i++)
		{
			if (depthRecvPtr[i]<myOwnDPtr[i])
			{
				memcpy(myOwnCPtr+i*4, colorRecvPtr+i*4, 4);
				//myOwnCPtr[i]=colorRecvPtr[i];
				myOwnDPtr[i]=depthRecvPtr[i];
			}
		}

		DEBUG_INFO("sending binary swap\n");

		SAFE_CALL(MPI_Wait(depthBufferExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));
		SAFE_CALL(MPI_Wait(colorBufferExchangeRequests+myTesterIndex, MPI_STATUS_IGNORE));
		//debugging...	
		//memset(hisDepthBuffer, 0, DepthBufferSize*sizeof(float));
		//memset(hisVisibleTable, 0, VisibleTableSize);
	}
	while (binarySwapDescriptor->next());
}



//assuming TesterCount, TileCount are all power of 2
void TesterContext::sendColorBuffer()
{
	int i, j, tileIndex;
	ubyte* colorBufferSendPtr;

	int displayerIndex;
#ifdef	USING_SEPARATE_DISPLAYER
	displayerIndex=DisplayerIndex;
#else
	displayerIndex=MasterIndex;
#endif

	tileIndex=binarySwapDescriptor->myRecvIndex;

	assert(tileIndex>=0&&tileIndex<DepthBufferSnapShotNumber);


	colorBufferSendPtr=myColorBuffer+tileIndex*ColorBufferAtomSize;

	//assert(binarySwapDescriptor->myTileCount==1);
	for (i=0; i<binarySwapDescriptor->myTileCount; i++)
	{		
		SAFE_CALL(MPI_Isend(colorBufferSendPtr+i*ColorBufferAtomSize, ColorBufferAtomSize,
			MPI_UNSIGNED_CHAR, displayerIndex, tileIndex+i, MPI_COMM_WORLD, colorBufferExchangeRequests+tileIndex+i));
	}
	//printf("Tester%d sends visible pixelInfo of total %d\n", myTesterIndex, counter);
}



void TesterContext::testColorBufferSending()
{
	int i;
	//printf("*********wait for #[%u] starting from [%u]*******\n", binarySwapDescriptor->myTileCount, binarySwapDescriptor->myRecvIndex);
	SAFE_CALL(MPI_Waitall(binarySwapDescriptor->myTileCount, colorBufferExchangeRequests+binarySwapDescriptor->myRecvIndex, 
		MPI_STATUSES_IGNORE));

	//printf("###Tester%d finishes sending visible pixelInfo\n", myTesterIndex);
}







			
			












